import logging
from unittest.mock import Mock
from unittest.mock import call
from unittest.mock import patch

import pytest

from saltext.vault.utils import vault
from saltext.vault.utils.vault import cache as vcache
from saltext.vault.utils.vault import client as vclient
from saltext.vault.utils.vault import leases as vleases

pytestmark = [pytest.mark.usefixtures("time_stopped")]


@pytest.fixture
def lease_renewed_response():
    return {
        "lease_id": "database/creds/testrole/abcd",
        "renewable": True,
        "lease_duration": 2000,
    }


@pytest.fixture
def lease_renewed_extended_response():
    return {
        "lease_id": "database/creds/testrole/abcd",
        "renewable": True,
        "lease_duration": 3000,
    }


@pytest.fixture
def store(events):
    client = Mock(spec=vclient.AuthenticatedVaultClient)
    cache = Mock(spec=vcache.VaultLeaseCache)
    cache.exists.return_value = False
    cache.get.return_value = None
    return vleases.LeaseStore(client, cache, expire_events=events)


@pytest.fixture
def store_valid(store, lease, lease_renewed_response):
    store.cache.exists.return_value = True
    store.cache.get.return_value = vleases.VaultLease(**lease)
    store.client.post.return_value = lease_renewed_response
    return store


@pytest.fixture
def store_multi(store, lease, lease_renewed_response):
    lease = vleases.VaultLease(**lease)
    lease_2 = lease.with_renewed(id="foobar", lease_id="foobar")
    lease_3 = lease.with_renewed(id="barbaz", lease_id="barbaz")
    leases = {"test_1": lease, "test_12": lease_2, "test_3": lease_3}
    store.cache.exists.side_effect = lambda x, **y: x in leases
    store.cache.get.side_effect = lambda x, **y: leases[x]
    store.cache.list.return_value = list(leases)
    store.client.post.return_value = lease_renewed_response
    return store


@pytest.mark.parametrize(
    "creation_time",
    [
        1661188581,
        "1661188581",
        "2022-08-22T17:16:21.473219641+00:00",
        "2022-08-22T17:16:21.47321964+00:00",
        "2022-08-22T17:16:21.4732196+00:00",
        "2022-08-22T17:16:21.473219+00:00",
        "2022-08-22T17:16:21.47321+00:00",
        "2022-08-22T17:16:21.4732+00:00",
        "2022-08-22T17:16:21.473+00:00",
        "2022-08-22T17:16:21.47+00:00",
        "2022-08-22T17:16:21.4+00:00",
    ],
)
def test_vault_lease_creation_time_normalization(creation_time):
    """
    Ensure the normalization of different creation_time formats works as expected -
    many token endpoints report a timestamp, while other endpoints report RFC3339-formatted
    strings that may have a variable number of digits for sub-second precision (0 omitted)
    while datetime.fromisoformat expects exactly 6 digits
    """
    data = {
        "lease_id": "id",
        "renewable": False,
        "lease_duration": 1337,
        "creation_time": creation_time,
        "data": None,
    }
    res = vleases.VaultLease(**data)
    assert res.creation_time == 1661188581


@pytest.mark.parametrize(
    "time_stopped,duration,offset,expected",
    [
        (0, 50, 0, True),
        (50, 10, 0, False),
        (0, 60, 10, True),
        (0, 60, 600, False),
    ],
    indirect=["time_stopped"],
)
def test_vault_lease_is_valid_accounts_for_time(duration, offset, expected):
    """
    Ensure lease validity is checked correctly and can look into the future
    """
    data = {
        "lease_id": "id",
        "renewable": False,
        "lease_duration": duration,
        "creation_time": 0,
        "expire_time": duration,
        "data": None,
    }
    res = vleases.VaultLease(**data)
    assert res.is_valid_for(offset) is expected


@pytest.mark.parametrize(
    "time_stopped,duration,offset,expected",
    [
        (0, 50, 0, True),
        (50, 10, 0, False),
        (0, 60, 10, True),
        (0, 60, 600, False),
    ],
    indirect=["time_stopped"],
)
def test_vault_token_is_valid_accounts_for_time(duration, offset, expected):
    """
    Ensure token time validity is checked correctly and can look into the future
    """
    data = {
        "client_token": "id",
        "renewable": False,
        "lease_duration": duration,
        "num_uses": 0,
        "creation_time": 0,
        "expire_time": duration,
    }
    res = vault.VaultToken(**data)
    assert res.is_valid_for(offset) is expected


@pytest.mark.parametrize(
    "num_uses,uses,expected",
    [(0, 999999, True), (1, 0, True), (1, 1, False), (1, 2, False)],
)
def test_vault_token_is_valid_accounts_for_num_uses(num_uses, uses, expected):
    """
    Ensure token uses validity is checked correctly
    """
    data = {
        "client_token": "id",
        "renewable": False,
        "lease_duration": 0,
        "num_uses": num_uses,
        "creation_time": 0,
        "use_count": uses,
    }
    with patch(
        "saltext.vault.utils.vault.leases.BaseLease.is_valid_for",
        autospec=True,
        return_value=True,
    ):
        res = vault.VaultToken(**data)
        assert res.is_valid() is expected


@pytest.mark.parametrize(
    "time_stopped,duration,offset,expected",
    [
        (0, 50, 0, True),
        (50, 10, 0, False),
        (0, 60, 10, True),
        (0, 60, 600, False),
    ],
    indirect=["time_stopped"],
)
def test_vault_approle_secret_id_is_valid_accounts_for_time(duration, offset, expected):
    """
    Ensure secret ID time validity is checked correctly and can look into the future
    """
    data = {
        "secret_id": "test-secret-id",
        "renewable": False,
        "creation_time": 0,
        "expiration_time": duration,
        "secret_id_num_uses": 0,
        "secret_id_ttl": duration,
    }
    res = vault.VaultSecretId(**data)
    assert res.is_valid(offset) is expected


@pytest.mark.parametrize(
    "num_uses,uses,expected",
    [(0, 999999, True), (1, 0, True), (1, 1, False), (1, 2, False)],
)
def test_vault_approle_secret_id_is_valid_accounts_for_num_uses(num_uses, uses, expected):
    """
    Ensure secret ID uses validity is checked correctly
    """
    data = {
        "secret_id": "test-secret-id",
        "renewable": False,
        "creation_time": 0,
        "secret_id_ttl": 0,
        "secret_id_num_uses": num_uses,
        "use_count": uses,
    }
    with patch(
        "saltext.vault.utils.vault.leases.BaseLease.is_valid_for",
        autospec=True,
        return_value=True,
    ):
        res = vault.VaultSecretId(**data)
        assert res.is_valid() is expected


class TestLeaseStore:
    """
    Vault Lease Tests
    """

    def test_get_uncached_or_invalid(self, store):
        """
        Ensure uncached or invalid leases are reported as None.
        """
        ret = store.get("test")
        assert ret is None
        store.client.post.assert_not_called()
        store.cache.flush.assert_not_called()
        store.cache.store.assert_not_called()

    def test_get_cached_valid(self, store_valid, lease):
        """
        Ensure valid leases are returned without extra behavior.
        """
        ret = store_valid.get("test")
        assert ret == lease
        store_valid.client.post.assert_not_called()
        store_valid.cache.flush.assert_not_called()
        store_valid.cache.store.assert_not_called()

    @pytest.mark.parametrize("valid_for", [2000, pytest.param(2002, id="2002_renewal_leeway")])
    def test_get_valid_renew_default_period(self, store_valid, lease, valid_for):
        """
        Ensure renewals are attempted by default, the cache is updated accordingly
        and validity checks after renewal allow for a little leeway to account
        for latency.
        """
        ret = store_valid.get("test", valid_for=valid_for)
        lease["duration"] = lease["expire_time"] = 2000
        assert ret == lease
        store_valid.client.post.assert_called_once_with(
            "sys/leases/renew", payload={"lease_id": lease["id"]}
        )
        store_valid.cache.flush.assert_not_called()
        store_valid.cache.store.assert_called_once_with("test", ret)
        store_valid.expire_events.assert_not_called()

    @pytest.mark.parametrize("lease", ({"min_ttl": 3000},), indirect=True)
    def test_get_valid_renew_min_ttl_on_lease(
        self, store_valid, lease, lease_renewed_extended_response
    ):
        """
        Ensure that even if the cached lease fulfills the requested valid_for,
        that a renewal is attempted if min_ttl is higher.
        """
        store_valid.client.post.return_value = lease_renewed_extended_response
        ret = store_valid.get("test", valid_for=1200)
        lease["duration"] = lease["expire_time"] = 3000
        assert ret == lease
        store_valid.client.post.assert_called_once_with(
            "sys/leases/renew", payload={"lease_id": lease["id"]}
        )
        store_valid.cache.flush.assert_not_called()
        store_valid.cache.store.assert_called_once_with("test", ret)
        store_valid.expire_events.assert_not_called()

    @pytest.mark.parametrize("lease", ({"min_ttl": 3000},), indirect=True)
    def test_get_valid_renew_min_ttl_on_lease_undercut(
        self, store_valid, lease, lease_renewed_response
    ):
        """
        Ensure that even if the cached lease fulfills the requested valid_for
        and a renewal still undercuts it, nothing is returned
        """
        store_valid.client.post.return_value = lease_renewed_response
        ret = store_valid.get("test", valid_for=1200)
        assert ret is None
        store_valid.client.post.assert_has_calls(
            (
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"]},
                ),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"], "increment": 3000},
                ),
            )
        )
        store_valid.cache.flush.assert_called_once_with("test")
        store_valid.expire_events.assert_called_once_with(
            tag="vault/lease/test/expire",
            data={"valid_for_less": 3000, "ttl": 2000, "meta": lease["meta"]},
        )

    @pytest.mark.parametrize("lease", ({"renew_increment": 2005},), indirect=True)
    def test_get_valid_renew_lease_default_period(self, store_valid, lease):
        """
        Ensure that renewals without increment override respect the value
        set on the lease during creation.
        """
        ret = store_valid.get("test", valid_for=2000)
        lease["duration"] = lease["expire_time"] = 2000
        assert ret == lease
        store_valid.client.post.assert_called_once_with(
            "sys/leases/renew",
            payload={"lease_id": lease["id"], "increment": lease["renew_increment"]},
        )
        store_valid.cache.flush.assert_not_called()
        store_valid.cache.store.assert_called_once_with("test", ret)
        store_valid.expire_events.assert_not_called()

    @pytest.mark.parametrize(
        "lease", ({}, {"renew_increment": 1999}, {"renew_increment": 2001}), indirect=True
    )
    def test_get_valid_renew_increment(self, store_valid, lease):
        """
        Ensure renew_increment is honored when renewing and overrides the one
        set on the lease only if it does not undercut it.
        """
        ret = store_valid.get("test", valid_for=1400, renew_increment=2000)
        lease["duration"] = lease["expire_time"] = 2000
        assert ret == lease
        store_valid.client.post.assert_called_once_with(
            "sys/leases/renew",
            payload={
                "lease_id": lease["id"],
                "increment": max(lease["renew_increment"] or 0, 2000),
            },
        )
        store_valid.cache.flush.assert_not_called()
        store_valid.cache.store.assert_called_once_with("test", ret)
        store_valid.expire_events.assert_not_called()

    def test_get_valid_renew_increment_insufficient(self, store_valid, lease):
        """
        Ensure that when renew_increment is set, valid_for is respected and that
        a second renewal using valid_for as increment is not attempted when the
        Vault server does not allow renewals for at least valid_for.
        If an event factory was passed, an event should be sent.
        """
        ret = store_valid.get("test", valid_for=2100, renew_increment=3000)
        assert ret is None
        store_valid.client.post.assert_has_calls(
            (
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"], "increment": 3000},
                ),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"], "increment": 60},
                ),
            )
        )
        store_valid.cache.flush.assert_called_once_with("test")
        store_valid.expire_events.assert_called_once_with(
            tag="vault/lease/test/expire", data={"valid_for_less": 2100, "ttl": 2000, "meta": None}
        )

    @pytest.mark.parametrize("lease", ({"min_ttl": 2100},), indirect=True)
    def test_get_valid_renew_increment_undercuts_min_ttl(
        self, store_valid, lease, caplog, lease_renewed_response, lease_renewed_extended_response
    ):
        """
        Ensure that when a renew_increment undercuts the lease's default
        min_ttl, the renewal is first attempted without a renew_increment
        and if it still undercuts the min_ttl, a second one is attempted
        with the min_ttl as the increment.
        The user should be warned about the invalid request.
        """
        store_valid.client.post.side_effect = (
            lease_renewed_response,
            lease_renewed_extended_response,
        )
        with caplog.at_level(logging.WARNING):
            ret = store_valid.get("test", valid_for=1500, renew_increment=1600)
            assert "Dropping requested renew_increment" in caplog.text
        lease["duration"] = lease["expire_time"] = 3000
        assert ret == lease

        store_valid.client.post.assert_has_calls(
            (
                call("sys/leases/renew", payload={"lease_id": lease["id"]}),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"], "increment": lease["min_ttl"]},
                ),
            )
        )
        store_valid.cache.flush.assert_not_called()
        store_valid.cache.store.assert_called_with("test", ret)
        store_valid.expire_events.assert_not_called()

    @pytest.mark.parametrize("valid_for", [3000, pytest.param(3002, id="3002_renewal_leeway")])
    def test_get_valid_renew_valid_for(
        self,
        store_valid,
        lease,
        valid_for,
        lease_renewed_response,
        lease_renewed_extended_response,
    ):
        """
        Ensure that, if renew_increment was not set and the default period
        does not yield valid_for, a second renewal is attempted by valid_for.
        There should be some leeway by default to account for latency.
        """
        store_valid.client.post.side_effect = (
            lease_renewed_response,
            lease_renewed_extended_response,
        )
        ret = store_valid.get("test", valid_for=valid_for)
        lease["duration"] = lease["expire_time"] = 3000
        assert ret == lease
        store_valid.client.post.assert_has_calls(
            (
                call("sys/leases/renew", payload={"lease_id": lease["id"]}),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"], "increment": valid_for},
                ),
            )
        )
        store_valid.cache.flush.assert_not_called()
        store_valid.cache.store.assert_called_with("test", ret)
        store_valid.expire_events.assert_not_called()

    @pytest.mark.parametrize("lease", ({}, {"meta": {"foo": "bar"}}), indirect=True)
    def test_get_valid_not_renew(self, store_valid, lease):
        """
        Currently valid leases should not be returned if they undercut
        valid_for. By default, revocation should be attempted and the cache
        should be flushed. If an event factory was passed, an event should be sent
        which includes the per-lease metadata.
        """
        ret = store_valid.get("test", valid_for=2000, renew=False)
        assert ret is None
        store_valid.cache.store.assert_not_called()
        store_valid.client.post.assert_called_once_with(
            "sys/leases/renew", payload={"lease_id": lease["id"], "increment": 60}
        )
        store_valid.cache.flush.assert_called_once_with("test")
        store_valid.expire_events.assert_called_once_with(
            tag="vault/lease/test/expire",
            data={"valid_for_less": 2000, "ttl": 1337, "meta": lease["meta"]},
        )

    @pytest.mark.parametrize(
        "lease", ({}, {"revoke_delay": 1200}, {"revoke_delay": 1400}), indirect=True
    )
    @pytest.mark.parametrize("revoke", (None, 1100, 1500))
    def test_get_valid_revoke(self, store_valid, lease, revoke):
        """
        Ensure revoke is honored when revoking and overrides the revoke_delay
        set on the lease. Also ensure the revoke_delay on the lease is
        the default.
        """
        ret = store_valid.get("test", valid_for=3500, revoke=revoke)
        assert ret is None
        store_valid.client.post.assert_called_with(
            "sys/leases/renew",
            payload={"lease_id": lease["id"], "increment": revoke or lease["revoke_delay"] or 60},
        )
        store_valid.cache.flush.assert_called_once_with("test")
        store_valid.expire_events.assert_called_once_with(
            tag="vault/lease/test/expire",
            data={"valid_for_less": 3500, "ttl": 2000, "meta": lease["meta"]},
        )

    def test_get_valid_not_flush(self, store_valid):
        """
        Currently valid leases should not be returned if they undercut
        valid_for and should not be revoked if requested so.
        If an event factory was passed, an event should be sent.
        """
        ret = store_valid.get("test", valid_for=2000, revoke=False, renew=False)
        assert ret is None
        store_valid.cache.flush.assert_not_called()
        store_valid.client.post.assert_not_called()
        store_valid.cache.store.assert_not_called()
        store_valid.expire_events.assert_called_once_with(
            tag="vault/lease/test/expire", data={"valid_for_less": 2000, "ttl": 1337, "meta": None}
        )

    @pytest.mark.parametrize("check_server", (False, True))
    @pytest.mark.parametrize("expected", (False, True))
    def test_get_check_server(
        self,
        store_valid,
        lease,
        expected,
        check_server,
        lease_lookup_response,
        lease_renewed_response,
    ):
        """
        Ensure that "valid" leases are validated with the server only if requested.
        """
        if expected:
            store_valid.client.put.return_value = lease_lookup_response
        else:
            store_valid.client.put.side_effect = (
                vault.VaultInvocationError("invalid lease"),
                lease_renewed_response,
            )
        ret = store_valid.get("test", valid_for=1300, check_server=check_server)
        assert (ret == lease) is (expected or not check_server)
        if check_server:
            if expected:
                store_valid.client.put.assert_called_once_with(
                    "sys/leases/lookup", payload={"lease_id": lease["id"]}
                )
                store_valid.cache.flush.assert_not_called()
                store_valid.expire_events.assert_not_called()
            else:
                store_valid.client.put.assert_called_once_with(
                    "sys/leases/lookup", payload={"lease_id": lease["id"]}
                )
                store_valid.client.post.assert_called_once_with(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"], "increment": 60},
                )
                store_valid.cache.flush.assert_called_once_with("test")
                store_valid.expire_events.assert_called_once_with(
                    tag="vault/lease/test/expire",
                    data={"valid_for_less": 1300, "ttl": 0, "meta": lease["meta"]},
                )
        else:
            store_valid.client.put.assert_not_called()
            store_valid.expire_events.assert_not_called()
        store_valid.cache.store.assert_not_called()

    @pytest.mark.parametrize("on_call", (1, 2))
    def test_get_renew_already_revoked(self, store_valid, on_call, lease_renewed_response):
        """
        Ensure there's no crash when a renewal is attempted on a revoked lease
        and that the cache is flushed.
        """
        store_valid.client.post.side_effect = (
            (on_call - 1) * [lease_renewed_response]
            + [vault.VaultInvocationError("lease not found")]
            + [lease_renewed_response]
        )
        ret = store_valid.get("test", valid_for=3000)
        assert ret is None
        store_valid.cache.flush.assert_called_once_with("test")
        store_valid.expire_events.assert_called_once_with(
            tag="vault/lease/test/expire", data={"valid_for_less": 3000, "ttl": 0, "meta": None}
        )

    @pytest.mark.parametrize("as_str", (False, True))
    def test_revoke_already_revoked(self, store_valid, lease, as_str):
        """
        A notfound exception should not matter for revoking.
        """
        store_valid.client.post.side_effect = vault.VaultInvocationError("lease not found")
        lease = vault.VaultLease(**lease)
        if as_str:
            lease = str(lease)
        assert store_valid.revoke(lease) is True

    def test_revoke_delay_should_have_a_minimum_of_1(self, store_valid, lease):
        """
        Since revocation internally uses renewals by setting the lease
        validity to a low value, a minimum value of 1 must be enforced,
        otherwise the lease is renewed by its default TTL.
        """
        assert store_valid.revoke(lease["id"], delta=0) is True
        store_valid.client.post.assert_called_once_with(
            "sys/leases/renew", payload={"lease_id": lease["id"], "increment": 1}
        )

    def test_list_info(self, store_multi, lease):
        """
        Test listing cached leases.
        """
        ret = store_multi.list_info()
        assert set(ret) == {"test_1", "test_12", "test_3"}
        lease.pop("data")
        lease["expires_in"] = 1337
        lease["expired"] = False
        assert ret["test_1"] == lease
        assert ret["test_12"]["lease_id"] == "foobar"
        assert ret["test_3"]["lease_id"] == "barbaz"

    def test_list_info_match(self, store_multi, lease):
        """
        Test listing cached leases with a glob.
        """
        ret = store_multi.list_info(match="test_1*")
        assert set(ret) == {"test_1", "test_12"}
        lease.pop("data")
        lease["expires_in"] = 1337
        lease["expired"] = False
        assert ret["test_1"] == lease
        assert ret["test_12"]["lease_id"] == "foobar"

    def test_list_info_expired(self, store_multi, lease):
        """
        The list should not contain expired leases, but they should
        not be flushed from cache during the call.
        """
        prev = store_multi.cache.get.side_effect
        store_multi.cache.get.side_effect = lambda x, **y: prev(x, **y) if x != "test_12" else None
        ret = store_multi.list_info(match="test_1*")
        assert set(ret) == {"test_1"}
        lease.pop("data")
        lease["expires_in"] = 1337
        lease["expired"] = False
        assert ret["test_1"] == lease
        store_multi.cache.get.assert_called_with("test_12", flush=False)

    def test_renew_cached(self, store_multi, lease):
        """
        Test renewing all cached leases.
        """
        assert store_multi.renew_cached() is True
        store_multi.client.post.assert_has_calls(
            (
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"]},
                ),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": "foobar"},
                ),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": "barbaz"},
                ),
            )
        )

    def test_renew_cached_match(self, store_multi, lease):
        """
        Test renewing all cached leases that match a glob.
        """
        assert store_multi.renew_cached(match="test_1*") is True
        store_multi.client.post.assert_has_calls(
            (
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"]},
                ),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": "foobar"},
                ),
            )
        )

    def test_renew_cached_expired(self, store_multi, lease):
        """
        Expired leases should be skipped and they should be flushed
        from cache during the call.
        """
        prev = store_multi.cache.get.side_effect
        store_multi.cache.get.side_effect = lambda x, **y: prev(x, **y) if x != "test_12" else None
        assert store_multi.renew_cached(match="test_1*") is True
        store_multi.client.post.assert_called_once_with(
            "sys/leases/renew", payload={"lease_id": lease["id"]}
        )
        store_multi.cache.get.assert_called_with("test_12", flush=True)

    @pytest.mark.parametrize("exc", (vault.VaultNotFoundError, vault.VaultPermissionDeniedError))
    def test_renew_cached_exception(self, store_multi, exc, lease_renewed_response, lease):
        """
        Ensure that exceptions are reraised after everything has been processed.
        """
        store_multi.client.post.side_effect = [lease_renewed_response, exc, lease_renewed_response]
        with pytest.raises(
            vault.VaultException, match=r"Failed renewing some leases: \['test_12'\]"
        ):
            store_multi.renew_cached()
        store_multi.client.post.assert_has_calls(
            (
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"]},
                ),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": "foobar"},
                ),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": "barbaz"},
                ),
            )
        )

    def test_revoke_cached(self, store_multi, lease):
        """
        Test revoking all cached leases.
        """
        assert store_multi.revoke_cached() is True
        store_multi.client.post.assert_has_calls(
            (
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"], "increment": 60},
                ),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": "foobar", "increment": 60},
                ),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": "barbaz", "increment": 60},
                ),
            )
        )

    def test_revoke_cached_match(self, store_multi, lease):
        """
        Test revoking all cached leases that match a glob.
        """
        assert store_multi.revoke_cached(match="test_1*", delta=120) is True
        store_multi.client.post.assert_has_calls(
            (
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"], "increment": 120},
                ),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": "foobar", "increment": 120},
                ),
            )
        )

    def test_revoke_cached_expired(self, store_multi, lease):
        """
        Expired leases should be skipped and they should be flushed
        from cache during the call.
        """
        prev = store_multi.cache.get.side_effect
        store_multi.cache.get.side_effect = lambda x, **y: prev(x, **y) if x != "test_12" else None
        assert store_multi.revoke_cached(match="test_1*") is True
        store_multi.client.post.assert_called_once_with(
            "sys/leases/renew", payload={"lease_id": lease["id"], "increment": 60}
        )
        store_multi.cache.get.assert_called_with("test_12", flush=True)

    @pytest.mark.parametrize("flush_on_failure", (False, True))
    def test_revoke_cached_exception(self, store_multi, lease_renewed_response, flush_on_failure):
        """
        Ensure that exceptions are reraised after everything has been processed.
        The cached leases should only be flushed from cache if it was requested.
        """
        store_multi.client.post.side_effect = [
            lease_renewed_response,
            vault.VaultPermissionDeniedError,
            lease_renewed_response,
        ]
        with pytest.raises(
            vault.VaultException, match=r"Failed revoking some leases: \['test_12'\]"
        ):
            store_multi.revoke_cached(flush_on_failure=flush_on_failure)
        if flush_on_failure:
            store_multi.cache.flush.assert_has_calls(
                (call("test_1"), call("test_12"), call("test_3"))
            )
        else:
            store_multi.cache.flush.assert_has_calls((call("test_1"), call("test_3")))



import os
import sys
from functools import wraps
from pathlib import Path

from . import prompt

try:
    # In case we have it, use it.
    # It's always installed in the Copier environment, so if you ensure you
    # call this via ``copier_python``, this will work.
    import yaml
except ImportError:
    yaml = None


if os.environ.get("STAGE"):
    # If we're running inside a Copier task/migration, cwd is the target dir.
    # We cannot use __file__ because this file is imported from the template clone.
    COPIER_ANSWERS = Path(".copier-answers.yml").resolve()
else:
    COPIER_ANSWERS = (Path(__file__).parent.parent.parent / ".copier-answers.yml").resolve()


if yaml is not None:

    def represent_str(dumper, data):
        """
        Represent multiline strings using "|"
        """
        if len(data.splitlines()) > 1:
            return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
        return dumper.represent_scalar("tag:yaml.org,2002:str", data)

    class OpinionatedYamlDumper(yaml.SafeDumper):
        """
        Indent lists by two spaces
        """

        def increase_indent(self, flow=False, indentless=False):
            return super().increase_indent(flow=flow, indentless=False)

    OpinionatedYamlDumper.add_representer(str, represent_str)


def _needs_answers(func):
    @wraps(func)
    def _wrapper(*args, **kwargs):
        if not COPIER_ANSWERS.exists():
            raise RuntimeError(f"Missing answers file at {COPIER_ANSWERS}")
        return func(*args, **kwargs)

    return _wrapper


@_needs_answers
def load_answers():
    """
    Load the complete answers file. Depends on PyYAML.
    """
    if not yaml:
        raise RuntimeError("Missing pyyaml in environment")
    with open(COPIER_ANSWERS, encoding="utf-8") as f:
        return yaml.safe_load(f)


@_needs_answers
def dump_answers(answers):
    """
    Write the complete answers file. Depends on PyYAML.
    Intended for answers migrations.
    """
    if not yaml:
        raise RuntimeError("Missing pyyaml in environment")
    with open(COPIER_ANSWERS, "w", encoding="utf-8") as f:
        yaml.dump(
            answers,
            f,
            Dumper=OpinionatedYamlDumper,
            indent=0,
            default_flow_style=False,
            canonical=False,
        )


@_needs_answers
def discover_project_name():
    """
    Specifically discover project name. No dependency.
    """
    for line in COPIER_ANSWERS.read_text().splitlines():
        if line.startswith("project_name"):
            return line.split(":", maxsplit=1)[1].strip()
    raise RuntimeError("Failed discovering project name")


def finish_task(msg, success, err_exit=False, extra=None):
    """
    Print final conclusion of task (migration) run in Copier.

    We usually want to exit with 0, even when something fails,
    because a failing task/migration should not crash Copier.
    """
    prompt.pprint("\n", stream=sys.stderr)
    if success:
        prompt.pprint(f"\n    ✓ {msg}", bold=True, bg=prompt.DARKGREEN, stream=sys.stderr)
    elif success is None:
        prompt.pprint(
            f"\n    ✓ {msg}", bold=True, fg=prompt.YELLOW, bg=prompt.DARKGREEN, stream=sys.stderr
        )
        success = True
    else:
        prompt.warn(f"    ✗ {msg}", extra)
    raise SystemExit(int(not success and err_exit))



import sys

from helpers import prompt
from helpers.copier import finish_task
from helpers.git import ensure_git
from helpers.venv import ensure_project_venv

if __name__ == "__main__":
    args = sys.argv[1:]
    try:
        prompt.ensure_utf8()
        ensure_git()
        venv = ensure_project_venv(install_extras="--extras" in args)
    except Exception as err:  # pylint: disable=broad-except
        finish_task(
            f"Failed initializing environment: {err}",
            False,
            True,
            extra=(
                "No worries, just follow the manual steps documented here: "
                "https://salt-extensions.github.io/salt-extension-copier/topics/creation.html#first-steps"
            ),
        )
    if "--print-venv" in args:
        print(venv)
    finish_task("Successfully initialized environment", True)



from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import patch

import pytest
import requests
import salt.modules.event

from saltext.vault.utils import vault
from saltext.vault.utils.vault import auth as vauth
from saltext.vault.utils.vault import client as vclient
from saltext.vault.utils.vault import helpers as hlp


@pytest.fixture(params=[0])
def time_stopped(request):
    with patch(
        "time.time",
        autospec=True,
        return_value=request.param,
    ):
        yield


def _mock_json_response(data, status_code=200, reason=""):
    """
    Mock helper for http response
    """
    response = Mock(spec=requests.models.Response)
    response.json.return_value = data
    response.status_code = status_code
    response.reason = reason
    if status_code < 400:
        response.ok = True
    else:
        response.ok = False
        response.raise_for_status.side_effect = requests.exceptions.HTTPError
    return response


@pytest.fixture
def config_defaults():
    return {
        "auth": {
            "approle_mount": "approle",
            "approle_name": "salt-master",
            "method": "token",
            "secret_id": None,
            "token_lifecycle": {
                "minimum_ttl": 10,
                "renew_increment": None,
            },
        },
        "cache": {
            "backend": "session",
            "clear_attempt_revocation": 60,
            "clear_on_unauthorized": True,
            "config": 3600,
            "expire_events": False,
            "kv_metadata": "connection",
            "secret": "ttl",
        },
        "client": {
            "connect_timeout": vclient.DEFAULT_CONNECT_TIMEOUT,
            "read_timeout": vclient.DEFAULT_READ_TIMEOUT,
            "max_retries": vclient.DEFAULT_MAX_RETRIES,
            "backoff_factor": vclient.DEFAULT_BACKOFF_FACTOR,
            "backoff_max": vclient.DEFAULT_BACKOFF_MAX,
            "backoff_jitter": vclient.DEFAULT_BACKOFF_JITTER,
            "retry_post": vclient.DEFAULT_RETRY_POST,
            "retry_status": vclient.DEFAULT_RETRY_STATUS,
            "respect_retry_after": vclient.DEFAULT_RESPECT_RETRY_AFTER,
            "retry_after_max": vclient.DEFAULT_RETRY_AFTER_MAX,
        },
        "issue": {
            "allow_minion_override_params": False,
            "type": "token",
            "approle": {
                "mount": "salt-minions",
                "params": {
                    "bind_secret_id": True,
                    "secret_id_num_uses": 1,
                    "secret_id_ttl": 60,
                    "token_explicit_max_ttl": 60,
                    "token_num_uses": 10,
                },
            },
            "token": {
                "role_name": None,
                "params": {
                    "explicit_max_ttl": None,
                    "num_uses": 1,
                },
            },
            "wrap": "30s",
        },
        "issue_params": {},
        "metadata": {
            "entity": {
                "minion-id": "{minion}",
            },
            "token": {
                "saltstack-jid": "{jid}",
                "saltstack-minion": "{minion}",
                "saltstack-user": "{user}",
            },
        },
        "policies": {
            "assign": [
                "saltstack/minions",
                "saltstack/{minion}",
            ],
            "cache_time": 60,
            "refresh_pillar": None,
        },
        "server": {
            "namespace": None,
            "verify": None,
        },
    }


@pytest.fixture(params=[{}])
def server_config(request):
    conf = {
        "url": "http://127.0.0.1:8200",
        "namespace": None,
        "verify": None,
    }
    conf.update(request.param)
    return conf


@pytest.fixture(params=["token", "approle"])
def test_config(config_defaults, server_config, request):
    # avoid mutating default config
    defaults = config_defaults.copy()
    auth = defaults["auth"].copy()

    if request.param == "token":
        auth["token"] = "test-token"

    if request.param == "wrapped_token":
        auth["method"] = "wrapped_token"
        auth["token"] = "test-wrapped-token"

    if request.param == "approle":
        auth["method"] = "approle"
        auth["role_id"] = "test-role-id"
        auth["secret_id"] = "test-secret-id"

    if request.param == "approle_no_secretid":
        auth["method"] = "approle"
        auth["role_id"] = "test-role-id"
        auth["secret_id"] = False

    defaults["auth"] = auth
    defaults["server"] = server_config
    return defaults


@pytest.fixture(params=["token", "approle"])
def test_remote_config(config_defaults, server_config, request):
    # avoid mutating default config
    defaults = {
        "auth": config_defaults["auth"].copy(),
        "cache": config_defaults["cache"].copy(),
        "client": config_defaults["client"].copy(),
        "server": server_config,
    }

    if request.param == "token":
        defaults["auth"]["token"] = "test-token"

    if request.param == "wrapped_token":
        defaults["auth"]["method"] = "wrapped_token"
        defaults["auth"]["token"] = "test-wrapped-token"

    if request.param == "token_changed":
        defaults["auth"]["token"] = "test-token-changed"

    if request.param == "approle":
        defaults["auth"]["method"] = "approle"
        defaults["auth"]["role_id"] = "test-role-id"
        # actual remote config would not contain secret_id, but
        # this is used for testing both from local and from remote
        defaults["auth"]["secret_id"] = "test-secret-id"

    if request.param == "approle_no_secretid":
        defaults["auth"]["method"] = "approle"
        defaults["auth"]["role_id"] = "test-role-id"
        defaults["auth"]["secret_id"] = False

    # this happens when wrapped role_ids are merged by _query_master
    if request.param == "approle_wrapped_roleid":
        defaults["auth"]["method"] = "approle"
        defaults["auth"]["role_id"] = {"role_id": "test-role-id"}
        # actual remote config does not contain secret_id
        defaults["auth"]["secret_id"] = True
    return defaults


@pytest.fixture
def role_id_response():
    return {
        "request_id": "c85838c5-ecfe-6d07-4b28-1935ac2e304a",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": {"role_id": "58b4c650-3d13-5932-a2fa-03865c8e85d7"},
        "warnings": None,
    }


@pytest.fixture
def secret_id_response():
    return {
        "request_id": "c85838c5-ecfe-6d07-4b28-1935ac2e304a",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": {
            "secret_id_accessor": "84896a0c-1347-aa90-a4f6-aca8b7558780",
            "secret_id": "841771dc-11c9-bbc7-bcac-6a3945a69cd9",
            "secret_id_ttl": 1337,
        },
        "warnings": None,
    }


@pytest.fixture
def secret_id_meta_response():
    return {
        "request_id": "7c97d03d-2166-6217-8da1-19604febae5c",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": {
            "cidr_list": [],
            "creation_time": "2022-08-22T17:37:07.753989459+00:00",
            "expiration_time": "2339-07-13T13:23:46.753989459+00:00",
            "last_updated_time": "2022-08-22T17:37:07.753989459+00:00",
            "metadata": {},
            "secret_id_accessor": "b1c88755-f2f5-2fd2-4bcc-cade95f6ba96",
            "secret_id_num_uses": 0,
            "secret_id_ttl": 9999999999,
            "token_bound_cidrs": [],
        },
        "warnings": None,
    }


@pytest.fixture
def wrapped_role_id_response():
    return {
        "request_id": "",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": None,
        "warnings": None,
        "wrap_info": {
            "token": "test-wrapping-token",
            "accessor": "test-wrapping-token-accessor",
            "ttl": 180,
            "creation_time": "2022-09-10T13:37:12.123456789+00:00",
            "creation_path": "auth/approle/role/test-minion/role-id",
            "wrapped_accessor": "",
        },
    }


@pytest.fixture
def wrapped_secret_id_response():
    return {
        "request_id": "",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": None,
        "warnings": None,
        "wrap_info": {
            "token": "test-wrapping-token",
            "accessor": "test-wrapping-token-accessor",
            "ttl": 180,
            "creation_time": "2022-09-10T13:37:12.123456789+00:00",
            "creation_path": "auth/approle/role/test-minion/secret-id",
            "wrapped_accessor": "",
        },
    }


@pytest.fixture
def wrapped_role_id_lookup_response():
    return {
        "request_id": "31e7020e-3ce3-2c63-e453-d5da8a9890f1",
        "lease_id": "",
        "renewable": False,
        "lease_duration": 0,
        "data": {
            "creation_path": "auth/approle/role/test-minion/role-id",
            "creation_time": "2022-09-10T13:37:12.123456789+00:00",
            "creation_ttl": 180,
        },
        "wrap_info": None,
        "warnings": None,
        "auth": None,
    }


@pytest.fixture
def wrapped_token_auth_response():
    return {
        "request_id": "",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": None,
        "warnings": None,
        "wrap_info": {
            "token": "test-wrapping-token",
            "accessor": "test-wrapping-token-accessor",
            "ttl": 180,
            "creation_time": "2022-09-10T13:37:12.123456789+00:00",
            "creation_path": "auth/token/create/salt-minion",
            "wrapped_accessor": "",
        },
    }


@pytest.fixture
def token_lookup_self_response():
    return {
        "request_id": "0e8c388e-2cb6-bcb2-83b7-625127d568bb",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": {
            "accessor": "test-token-accessor",
            "creation_time": 1661188581,
            "creation_ttl": 9999999999,
            "display_name": "",
            "entity_id": "",
            "expire_time": "2339-07-13T11:03:00.473212541+00:00",
            "explicit_max_ttl": 0,
            "id": "test-token",
            "issue_time": "2022-08-22T17:16:21.473219641+00:00",
            "meta": {},
            "num_uses": 0,
            "orphan": True,
            "path": "",
            "policies": ["default"],
            "renewable": True,
            "ttl": 9999999999,
            "type": "service",
        },
        "warnings": None,
    }


@pytest.fixture
def token_renew_self_response():
    return {
        "auth": {
            "client_token": "test-token",
            "policies": ["default", "renewed"],
            "metadata": {},
        },
        "lease_duration": 3600,
        "renewable": True,
    }


@pytest.fixture
def token_renew_other_response():
    return {
        "auth": {
            "client_token": "other-test-token",
            "policies": ["default", "renewed"],
            "metadata": {},
        },
        "lease_duration": 3600,
        "renewable": True,
    }


@pytest.fixture
def token_renew_accessor_response():
    return {
        "auth": {
            "client_token": "",
            "policies": ["default", "renewed"],
            "metadata": {},
        },
        "lease_duration": 3600,
        "renewable": True,
    }


@pytest.fixture
def token_auth():
    return {
        "request_id": "0e8c388e-2cb6-bcb2-83b7-625127d568bb",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "auth": {
            "client_token": "test-token",
            "renewable": True,
            "lease_duration": 9999999999,
            "num_uses": 0,
            "creation_time": 1661188581,
        },
    }


@pytest.fixture
def lease_response():
    return {
        "request_id": "0e8c388e-2cb6-bcb2-83b7-625127d568bb",
        "lease_id": "database/creds/testrole/abcd",
        "lease_duration": 1337,
        "renewable": True,
        "data": {
            "username": "test",
            "password": "test",
        },
    }


@pytest.fixture
def lease(request):
    overrides = getattr(request, "param", {})
    base = {
        "id": "database/creds/testrole/abcd",
        "lease_id": "database/creds/testrole/abcd",
        "renewable": True,
        "duration": 1337,
        "creation_time": 0,
        "expire_time": 1337,
        "data": {
            "username": "test",
            "password": "test",
        },
        "meta": None,
        "min_ttl": None,
        "revoke_delay": None,
        "renew_increment": None,
    }
    base.update(overrides)
    return base


@pytest.fixture
def lease_lookup_response():
    return {
        "request_id": "852e6e12-7bf5-0f38-896d-4ced16c8c8f0",
        "lease_id": "",
        "renewable": False,
        "lease_duration": 0,
        "data": {
            "expire_time": 2000,
            "id": "database/creds/testrole/abcd",
            "issue_time": 0,
            "last_renewal": 0,
            "renewable": True,
            "ttl": 2000,
        },
        "wrap_info": None,
        "warnings": None,
        "auth": None,
    }


@pytest.fixture
def session():
    return Mock(spec=requests.Session)


@pytest.fixture
def req(session):
    yield session.request


@pytest.fixture
def req_failed(req, request):
    status_code = getattr(request, "param", 502)
    req.return_value = _mock_json_response({"errors": ["foo"]}, status_code=status_code)
    yield req


@pytest.fixture
def req_success(req):
    req.return_value = _mock_json_response(None, status_code=204)
    yield req


@pytest.fixture(params=[200])
def req_any(req, request):
    data = {}
    if request.param != 204:
        data["data"] = {"foo": "bar"}
    if request.param >= 400:
        data["errors"] = ["foo"]
    req.return_value = _mock_json_response(data, status_code=request.param)
    yield req


@pytest.fixture
def req_unwrapping(wrapped_role_id_lookup_response, role_id_response, req):
    req.side_effect = lambda method, url, **kwargs: (
        _mock_json_response(wrapped_role_id_lookup_response)
        if url.endswith("sys/wrapping/lookup")
        else _mock_json_response(role_id_response)
    )
    yield req


@pytest.fixture(params=["data"])
def unauthd_client_mock(server_config, request):
    client = Mock()
    client.get_config.return_value = server_config
    client.unwrap.return_value = {request.param: {"bar": "baz"}}
    yield client


@pytest.fixture(params=[None, "valid_token"])
def client(server_config, request, session):
    if request.param is None:
        return vclient.VaultClient(**server_config, session=session)
    if request.param == "valid_token":
        token = request.getfixturevalue(request.param)
        auth = Mock(spec=vauth.VaultTokenAuth)
        auth.is_renewable.return_value = True
        auth.is_valid.return_value = True
        auth.get_token.return_value = token
    if request.param == "invalid_token":
        token = request.getfixturevalue(request.param)
        auth = Mock(spec=vauth.VaultTokenAuth)
        auth.is_renewable.return_value = True
        auth.is_valid.return_value = False
        auth.get_token.side_effect = vault.VaultAuthExpired
    return vclient.AuthenticatedVaultClient(auth, **server_config, session=session)


@pytest.fixture
def valid_token(token_auth):
    token = MagicMock(spec=vault.VaultToken, **token_auth["auth"])
    token.is_valid.return_value = True
    token.is_renewable.return_value = True
    token.payload.return_value = {"token": token_auth["auth"]["client_token"]}
    token.__str__.return_value = token_auth["auth"]["client_token"]
    token.to_dict.return_value = token_auth["auth"]
    return token


@pytest.fixture
def invalid_token(valid_token):
    valid_token.is_valid.return_value = False
    valid_token.is_renewable.return_value = False
    return valid_token


@pytest.fixture
def cache_factory():
    with patch("salt.cache.factory", autospec=True) as factory:
        yield factory


@pytest.fixture
def events():
    return Mock(spec=salt.modules.event.send)


@pytest.fixture(params=["MASTER", "MASTER_IMPERSONATING", "MINION_LOCAL", "MINION_REMOTE"])
def salt_runtype(request):
    runtype = Mock(spec=hlp._get_salt_run_type)
    runtype.return_value = getattr(hlp, f"SALT_RUNTYPE_{request.param}")
    with patch("saltext.vault.utils.vault.helpers._get_salt_run_type", runtype):
        yield


@pytest.fixture(
    params=[
        "master",
        "master_impersonating",
        "minion_local_1",
        "minion_local_2",
        "minion_local_3",
        "minion_remote",
    ]
)
def opts_runtype(request):
    return {
        "master": {
            "__role": "master",
            "vault": {},
        },
        "master_peer_run": {
            "__role": "master",
            "grains": {
                "id": "test-minion",
            },
            "vault": {},
        },
        "master_impersonating": {
            "__role": "master",
            "minion_id": "test-minion",
            "grains": {
                "id": "test-minion",
            },
            "vault": {},
        },
        "minion_local_1": {
            "grains": {"id": "test-minion"},
            "local": True,
        },
        "minion_local_2": {
            "file_client": "local",
            "grains": {"id": "test-minion"},
        },
        "minion_local_3": {
            "grains": {"id": "test-minion"},
            "master_type": "disable",
        },
        "minion_remote": {
            "grains": {"id": "test-minion"},
        },
    }[request.param]



import pytest

from tests.support.vault import vault_read_secret
from tests.support.vault import vault_write_secret

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def minion_config_overrides(vault_port):
    return {
        "vault": {
            "auth": {
                "method": "token",
                "token": "testsecret",
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        }
    }


@pytest.fixture
def vault_secret(states):
    yield states.vault_secret


@pytest.fixture(params=(False, True))
def testmode(request):
    return request.param


@pytest.fixture
def temp_secret(modules):
    key = "secret/my/secret"
    yield key
    assert modules.vault.wipe_secret(key) is True


@pytest.fixture
def secret_present(temp_secret):
    vault_write_secret(temp_secret, foo="bar")
    assert vault_read_secret(temp_secret) == {"foo": "bar"}
    yield temp_secret


@pytest.mark.parametrize("sync", (False, True))
def test_present_create(vault_secret, temp_secret, sync, testmode):
    values = {"foo": "bar"}
    ret = vault_secret.present(temp_secret, values=values, sync=sync, test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("Would have" in ret.comment) is testmode
    res = vault_read_secret(temp_secret)
    assert (res != values) is testmode


@pytest.mark.parametrize("sync", (False, True))
def test_present_already_present(vault_secret, secret_present, sync, testmode):
    values = {"foo": "bar"}
    ret = vault_secret.present(secret_present, values=values, sync=sync, test=testmode)
    assert ret.result is True
    assert "as specified" in ret.comment
    assert not ret.changes
    res = vault_read_secret(secret_present)
    assert res == values


@pytest.mark.parametrize("sync", (False, True))
def test_present_change(vault_secret, secret_present, sync, testmode):
    values = {"bar": "baz"}
    ret = vault_secret.present(secret_present, values=values, sync=sync, test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("Would have" in ret.comment) is testmode
    assert ret.changes
    assert ("written" in ret.changes) is sync
    assert ("patched" in ret.changes) is not sync
    assert ret.changes[next(iter(ret.changes))] == secret_present
    res = vault_read_secret(secret_present)
    assert (res == values) is (not testmode and sync)
    assert ("foo" in res) is (testmode or not sync)


def test_present_change_patch(vault_secret, secret_present, testmode):
    values = {"foo": None, "bar": "baz"}
    ret = vault_secret.present(secret_present, values=values, sync=False, test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("Would have" in ret.comment) is testmode
    assert ret.changes
    assert "patched" in ret.changes
    assert ret.changes["patched"] == secret_present
    res = vault_read_secret(secret_present)
    assert ("foo" in res) is testmode
    assert (res != {"bar": "baz"}) is testmode


def test_absent_already_absent(vault_secret, testmode):
    ret = vault_secret.absent("secret/foo/bar/nonexistent", test=testmode)
    assert ret.result is True
    assert "already absent" in ret.comment
    assert not ret.changes


def test_absent(vault_secret, secret_present, testmode, modules):
    ret = vault_secret.absent(secret_present, test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("Would have" in ret.comment) is testmode
    assert ret.changes
    assert "deleted" in ret.changes
    assert ret.changes["deleted"] == secret_present
    res = vault_read_secret(secret_present)
    assert (res is None) is not testmode
    meta = modules.vault.read_secret_meta(secret_present)
    assert bool(meta["versions"]["1"]["deletion_time"]) is not testmode


def test_absent_destroy(vault_secret, secret_present, testmode, modules):
    ret = vault_secret.absent(secret_present, operation="destroy", test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("Would have" in ret.comment) is testmode
    assert ret.changes
    assert "destroyed" in ret.changes
    assert ret.changes["destroyed"] == secret_present
    res = vault_read_secret(secret_present)
    assert (res is None) is not testmode
    meta = modules.vault.read_secret_meta(secret_present)
    assert bool(meta["versions"]["1"]["destroyed"]) is not testmode


def test_absent_wipe(vault_secret, secret_present, testmode, modules):
    ret = vault_secret.absent(secret_present, operation="wipe", test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("Would have" in ret.comment) is testmode
    assert ret.changes
    assert "wiped" in ret.changes
    assert ret.changes["wiped"] == secret_present
    res = vault_read_secret(secret_present)
    assert (res is None) is not testmode
    meta = modules.vault.read_secret_meta(secret_present)
    assert (meta is False) is not testmode



import logging
import os
import shutil
import subprocess
import time
from pathlib import Path

import pytest
import salt.utils.path
import salt.utils.platform
from pytestshellutils.utils import ports
from pytestshellutils.utils.processes import ProcessResult
from saltfactories.utils import random_string

from saltext.vault import PACKAGE_ROOT
from tests.support.helpers import PatchedEnviron
from tests.support.vault import vault_enable_auth_method
from tests.support.vault import vault_enable_secret_engine
from tests.support.vault import vault_write_policy_file

try:
    import pwd
except ImportError:  # pragma: no cover
    import salt.utils.win_functions

# Reset the root logger to its default level(because salt changed it)
logging.root.setLevel(logging.WARNING)


# This swallows all logging to stdout.
# To show select logs, set --log-cli-level=<level>
for handler in logging.root.handlers[:]:  # pragma: no cover
    logging.root.removeHandler(handler)
    handler.close()

log = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def salt_factories_config():  # pragma: no cover
    """
    Return a dictionary with the keyword arguments for FactoriesManager
    """
    return {
        "code_dir": str(PACKAGE_ROOT),
        "inject_sitecustomize": "COVERAGE_PROCESS_START" in os.environ,
        "start_timeout": 120 if os.environ.get("CI") else 60,
    }


@pytest.fixture(scope="module")
def master_config_defaults(vault_port):  # pragma: no cover
    """
    This default configuration ensures the master issues authentication
    credentials with the correct policies. By default, it will issue
    tokens with an unlimited number of uses.
    """
    return {
        "peer_run": {
            ".*": [
                "vault.get_config",
                "vault.generate_new_token",
                "vault.generate_secret_id",
            ]
        },
        "sdbvault": {
            "driver": "vault",
        },
        "vault": {
            "auth": {
                "method": "token",
                "token": "testsecret",
            },
            "issue": {
                "token": {
                    "params": {
                        "num_uses": 0,
                    }
                }
            },
            "policies": {
                "assign": [
                    "salt_minion",
                ]
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        },
    }


@pytest.fixture(scope="module")
def master_config_overrides():  # pragma: no cover
    """
    You can override the default configuration per package by overriding this
    fixture in a conftest.py file.
    """
    return {}


@pytest.fixture(scope="module")
def master(salt_factories, master_config_defaults, master_config_overrides):  # pragma: no cover
    return salt_factories.salt_master_daemon(
        random_string("master-"), defaults=master_config_defaults, overrides=master_config_overrides
    )


@pytest.fixture(scope="module")
def minion_config_defaults(vault_port):  # pragma: no cover
    """
    The default minion configuration ensures that the minion works in --local
    mode and that the ``sdbvault`` SDB configuration is present.
    The vault configuration will not be used when not in masterless mode
    without overriding ``vault:config_location`` to ``local``.
    """
    return {
        "sdbvault": {
            "driver": "vault",
        },
        "vault": {
            "auth": {
                "method": "token",
                "token": "testsecret",
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        },
    }


@pytest.fixture(scope="module")
def minion_config_overrides():  # pragma: no cover
    """
    You can override the default configuration per package by overriding this
    fixture in a conftest.py file.
    """
    return {}


@pytest.fixture(scope="module")
def minion(master, minion_config_defaults, minion_config_overrides):  # pragma: no cover
    return master.salt_minion_daemon(
        random_string("minion-"), defaults=minion_config_defaults, overrides=minion_config_overrides
    )


@pytest.fixture(scope="session")
def current_user():  # pragma: no cover
    """
    Get the user associated with the current process.
    """
    if salt.utils.platform.is_windows():
        return salt.utils.win_functions.get_current_user(with_domain=False)
    return pwd.getpwuid(os.getuid())[0]


@pytest.fixture(scope="module")
def sshd_server(salt_factories, sshd_config_dir):  # pragma: no cover
    sshd_config_dict = {
        "Protocol": "2",
        # Turn strict modes off so that we can operate in /tmp
        "StrictModes": "no",
        # Logging
        "SyslogFacility": "AUTH",
        "LogLevel": "INFO",
        # Authentication:
        "LoginGraceTime": "120",
        "PermitRootLogin": "without-password",
        "PubkeyAuthentication": "yes",
        # Don't read the user's ~/.rhosts and ~/.shosts files
        "IgnoreRhosts": "yes",
        "HostbasedAuthentication": "no",
        # To enable empty passwords, change to yes (NOT RECOMMENDED)
        "PermitEmptyPasswords": "no",
        # Change to yes to enable challenge-response passwords (beware issues with
        # some PAM modules and threads)
        "ChallengeResponseAuthentication": "no",
        # Change to no to disable tunnelled clear text passwords
        "PasswordAuthentication": "no",
        "X11Forwarding": "no",
        "X11DisplayOffset": "10",
        "PrintMotd": "no",
        "PrintLastLog": "yes",
        "TCPKeepAlive": "yes",
        "AcceptEnv": "LANG LC_*",
        "UsePAM": "yes",
    }
    sftp_server_paths = [
        # Common
        "/usr/lib/openssh/sftp-server",
        # CentOS Stream 9
        "/usr/libexec/openssh/sftp-server",
        # Arch Linux
        "/usr/lib/ssh/sftp-server",
        # Photon OS 5
        "/usr/libexec/sftp-server",
    ]
    sftp_server_path = None
    for path in sftp_server_paths:
        if Path(path).exists():
            sftp_server_path = path
    if sftp_server_path is None:
        pytest.fail(f"Failed to find 'sftp-server'. Searched: {sftp_server_paths}")
    else:
        sshd_config_dict["Subsystem"] = f"sftp {sftp_server_path}"
    factory = salt_factories.get_sshd_daemon(
        sshd_config_dict=sshd_config_dict,
        config_dir=sshd_config_dir,
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="module")
def known_hosts_file(sshd_server, master, salt_factories):  # pragma: no cover
    with (
        pytest.helpers.temp_file(
            "ssh-known-hosts",
            "\n".join(sshd_server.get_host_keys()),
            salt_factories.tmp_root_dir,
        ) as known_hosts_file,
        pytest.helpers.temp_file(
            "master.d/ssh-known-hosts.conf",
            f"known_hosts_file: {known_hosts_file}",
            master.config_dir,
        ),
    ):
        yield known_hosts_file


@pytest.fixture(scope="module")
def salt_ssh_roster_file(
    sshd_server, master, known_hosts_file, current_user
):  # pylint: disable=unused-argument; pragma: no cover
    roster_contents = f"""
    localhost:
      host: 127.0.0.1
      port: {sshd_server.listen_port}
      user: {current_user}
    """
    if salt.utils.platform.is_darwin():
        roster_contents += "  set_path: $PATH:/usr/local/bin/\n"

    with pytest.helpers.temp_file("roster", roster_contents, master.config_dir) as roster_file:
        yield roster_file


@pytest.fixture(scope="session")
def sshd_config_dir(salt_factories):  # pragma: no cover
    config_dir = salt_factories.get_root_dir_for_daemon("sshd")
    try:
        yield config_dir
    finally:
        shutil.rmtree(str(config_dir), ignore_errors=True)


@pytest.fixture(scope="session")
def vault_port():
    return ports.get_unused_localhost_port()


@pytest.fixture(scope="module")
def vault_environ(vault_port):
    with PatchedEnviron(VAULT_ADDR=f"http://127.0.0.1:{vault_port}"):
        yield


CONTAINER_TARGETS = os.environ.get(
    "TESTING_CONTAINER", "hashicorp/vault:latest,openbao/openbao:latest"
).split(",")


@pytest.fixture(
    scope="module",
    params=CONTAINER_TARGETS,
)
def container(
    request, salt_factories, vault_port, vault_environ
):  # pylint: disable=unused-argument
    vault_binary = salt.utils.path.which("vault")

    if "openbao" in request.param:
        env = {
            "BAO_DEV_ROOT_TOKEN_ID": "testsecret",
        }
    else:
        env = {
            "VAULT_DEV_ROOT_TOKEN_ID": "testsecret",
        }

    factory = salt_factories.get_container(
        "vault",
        request.param,
        check_ports=[vault_port],
        container_run_kwargs={
            "ports": {"8200/tcp": vault_port},
            "environment": env,
        },
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
    )
    with factory.started() as factory:
        attempts = 0
        while attempts < 3:
            attempts += 1
            time.sleep(1)
            proc = subprocess.run(
                [vault_binary, "login", "token=testsecret"],
                check=False,
                capture_output=True,
                text=True,
            )
            if proc.returncode == 0:
                break
            ret = ProcessResult(
                returncode=proc.returncode,
                stdout=proc.stdout,
                stderr=proc.stderr,
                cmdline=proc.args,
            )
            log.debug("Failed to authenticate against vault:\n%s", ret)
            time.sleep(4)
        else:
            pytest.fail("Failed to login to vault")

        vault_write_policy_file("salt_master")
        vault_write_policy_file("salt_minion")
        vault_write_policy_file("database_admin")
        vault_write_policy_file("policy_admin")
        vault_write_policy_file("pki_admin")
        vault_write_policy_file("ssh_admin")

        vault_enable_auth_method("approle", ["-path=salt-minions"])
        vault_enable_secret_engine("kv", ["-version=1", "-path=secret-v1"])
        vault_enable_secret_engine("kv", ["-version=2", "-path=salt"])
        yield request.param


@pytest.fixture(scope="session")
def container_host_ref():
    # For Podman, there is `host.containers.internal`, which works even rootless.
    # This env var is set by nox.
    # `host.docker.internal` exists, but does not work in CI for some reason.
    # There, return the default IP address of the host on the default network (hardcoded).
    return os.environ.get("CONTAINER_HOST_REF", "172.17.0.1")



import json
import logging
import subprocess

import pytest
import salt.utils.files
import salt.utils.path
from pytestshellutils.utils.processes import ProcessResult

from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


def _vault_cmd(cmd, textinput=None, raw=False):
    vault_binary = salt.utils.path.which("vault")
    proc = subprocess.run(
        [vault_binary] + cmd,
        check=False,
        input=textinput,
        capture_output=True,
        text=True,
    )

    ret = ProcessResult(
        returncode=proc.returncode,
        stdout=proc.stdout,
        stderr=proc.stderr,
        cmdline=proc.args,
    )

    if raw:
        return ret
    if ret.returncode != 0:
        log.debug("Failed to run vault %s:\n%s\nSTDERR: %s", " ".join(cmd), ret, ret.stderr)
        raise RuntimeError(ret.stderr or ret.stdout)
    return ret


def vault_write_policy(name, rules):
    try:
        _vault_cmd(["policy", "write", name, "-"], textinput=rules)
    except RuntimeError as err:
        pytest.fail(f"Unable to write policy `{name}`: {err}")


def vault_write_policy_file(policy, filename=None):
    if filename is None:
        filename = policy
    try:
        _vault_cmd(
            [
                "policy",
                "write",
                policy,
                f"{RUNTIME_VARS.FILES}/vault/policies/{filename}.hcl",
            ]
        )
    except RuntimeError as err:
        pytest.fail(f"Unable to write policy `{policy}`: {err}")


def vault_read_policy(policy):
    ret = _vault_cmd(["policy", "read", "-format=json", policy], raw=True)
    if ret.returncode != 0:
        if "No policy named" in ret.stderr:
            return None
        log.debug("Failed to read policy `%s`:\n%s\nSTDERR: %s", policy, ret, ret.stderr)
        pytest.fail(f"Unable to read policy `{policy}`: {ret.stderr or ret.stdout}")
    res = json.loads(ret.stdout)
    return res["policy"]


def vault_list_policies():
    try:
        ret = _vault_cmd(["policy", "list", "-format=json"])
    except RuntimeError as err:
        pytest.fail(f"Unable to list policies: {err}")
    return json.loads(ret.stdout)


def vault_delete_policy(policy):
    try:
        _vault_cmd(["policy", "delete", policy])
    except RuntimeError as err:
        pytest.fail(f"Unable to delete policy `{policy}`: {err}")


def vault_enable_secret_engine(name, options=None):
    if options is None:
        options = []
    try:
        ret = _vault_cmd(["secrets", "enable"] + options + [name])
    except RuntimeError as err:
        pytest.fail(f"Could not enable secret engine `{name}`: {err}")

    if "path is already in use at" in ret.stdout:
        return False
    if "Success" in ret.stdout:
        return True
    log.debug("Failed to enable secret engine `%s`:\n%s\nSTDERR: %s", name, ret, ret.stderr)
    pytest.fail(f"Could not enable secret engine `{name}`: {ret.stderr or ret.stdout}")


def vault_disable_secret_engine(path):
    try:
        ret = _vault_cmd(["secrets", "disable", path])
    except RuntimeError as err:
        pytest.fail(f"Could not disable secret engine at `{path}`: {err}")

    if "Success" in ret.stdout:
        return True
    log.debug(
        "Failed to disable secret engine at path `%s`:\n%s\nSTDERR: %s", path, ret, ret.stderr
    )
    pytest.fail(f"Could not disable secret engine at path `{path}`: {ret.stderr or ret.stdout}")


def vault_enable_auth_method(name, options=None, **kwargs):
    if options is None:
        options = []
    cmd = ["auth", "enable"] + options + [name] + [f"{k}={v}" for k, v in kwargs.items()]
    try:
        ret = _vault_cmd(cmd)
    except RuntimeError as err:
        pytest.fail(f"Could not enable auth method `{name}`: {err}")

    if "path is already in use at" in ret.stdout:
        return False
    if "Success" in ret.stdout:
        return True
    log.debug("Failed to enable auth method `%s`:\n%s\nSTDERR: %s", name, ret, ret.stderr)
    pytest.fail(f"Could not enable auth method `{name}`: {ret.stderr or ret.stdout}")


def vault_disable_auth_method(name):
    try:
        ret = _vault_cmd(["auth", "disable", name])
    except RuntimeError as err:
        pytest.fail(f"Could not disable auth method `{name}`: {err}")

    if "Success" in ret.stdout:
        return True
    log.debug("Failed to disable auth method `%s`:\n%s\nSTDERR: %s", name, ret, ret.stderr)
    pytest.fail(f"Could not disable auth method `{name}`: {ret.stderr or ret.stdout}")


def vault_write_approle(name, mount="approle", **kwargs):
    cmd = ["write", "-f", f"auth/{mount}/role/{name}"] + [f"{k}={v}" for k, v in kwargs.items()]
    try:
        _vault_cmd(cmd)
    except RuntimeError as err:
        pytest.fail(f"Failed to write approle `{name}` at `{mount}`: {err}")


def vault_delete_approle(name, mount="approle"):
    cmd = ["delete", f"auth/{mount}/role/{name}"]
    try:
        _vault_cmd(cmd)
    except RuntimeError as err:
        pytest.fail(f"Failed to write approle `{name}` at `{mount}`: {err}")


def vault_get_role_id(name, mount="approle"):
    cmd = ["read", "-format=json", f"auth/{mount}/role/{name}/role-id"]
    try:
        ret = _vault_cmd(cmd)
    except RuntimeError as err:
        pytest.fail(f"Failed to read role-id for `{name}` at `{mount}`: {err}")
    return json.loads(ret.stdout)["data"]["role_id"]


def vault_create_secret_id(name, mount="approle"):
    cmd = ["write", "-f", "-format=json", f"auth/{mount}/role/{name}/secret-id"]
    try:
        ret = _vault_cmd(cmd)
    except RuntimeError as err:
        pytest.fail(f"Failed to create secret-id for `{name}` at `{mount}`: {err}")
    return json.loads(ret.stdout)["data"]["secret_id"]


def vault_write_secret(path, **kwargs):
    cmd = ["kv", "put", path] + [f"{k}={v}" for k, v in kwargs.items()]
    try:
        ret = _vault_cmd(cmd)
    except RuntimeError as err:
        pytest.fail(f"Failed to write secret at `{path}`: {err}")

    if vault_read_secret(path) != kwargs:
        log.debug("Failed to write secret at `%s`:\n%s\nSTDERR: %s", path, ret, ret.stderr)
        pytest.fail(f"Failed to write secret at `{path}`: {ret.stderr or ret.stdout}")
    return True


def vault_write_secret_file(path, data_name):
    data_path = f"{RUNTIME_VARS.FILES}/vault/data/{data_name}.json"
    with salt.utils.files.fopen(data_path) as f:
        data = json.load(f)
    cmd = ["kv", "put", path, f"@{data_path}"]
    try:
        ret = _vault_cmd([cmd])
    except RuntimeError as err:
        pytest.fail(f"Failed to write secret at `{path}`: {err}")

    if vault_read_secret(path) != data:
        log.debug("Failed to write secret at `%s`:\n%s\nSTDERR: %s", path, ret, ret.stderr)
        pytest.fail(f"Failed to write secret at `{path}`: {ret.stderr or ret.stdout}")
    return True


def vault_read_secret(path):
    ret = _vault_cmd(["kv", "get", "-format=json", path], raw=True)

    if ret.returncode != 0:
        if "No value found at" in ret.stderr:
            return None
        log.debug("Failed to read secret at `%s`:\n%s\nSTDERR: %s", path, ret, ret.stderr)
        pytest.fail(f"Failed to read secret at `{path}`: {ret.stderr or ret.stdout}")
    res = json.loads(ret.stdout)
    if "data" in res["data"]:
        return res["data"]["data"]
    return res["data"]


def vault_list_secrets(path):
    ret = _vault_cmd(["kv", "list", "-format=json", path], raw=True)
    if ret.returncode != 0:
        if ret.returncode == 2:
            return []
        log.debug("Failed to list secrets at `%s`:\n%s\nSTDERR: %s", path, ret, ret.stderr)
        pytest.fail(f"Failed to list secrets at `{path}`: {ret.stderr or ret.stdout}")
    return json.loads(ret.stdout)


def vault_delete_secret(path, metadata=False):
    try:
        ret = _vault_cmd(["kv", "delete", path])
    except RuntimeError:
        pytest.fail(f"Failed to delete secret at `{path}`")

    if vault_read_secret(path) is not None:
        log.debug("Failed to delete secret at `%s`:\n%s\nSTDERR: %s", path, ret, ret.stderr)
        pytest.fail(f"Failed to delete secret at `{path}`: {ret.stderr or ret.stdout}")

    if not metadata:
        return True

    ret = _vault_cmd(["kv", "metadata", "delete", path], raw=True)
    if ret.returncode != 0 and "Metadata not supported on KV Version 1" not in ret.stderr:
        log.debug(
            "Failed to delete secret metadata at `%s`:\n%s\nSTDERR: %s", path, ret, ret.stderr
        )
        pytest.fail(f"Failed to delete secret metadata at `{path}`: {ret.stderr or ret.stdout}")
    return True


def vault_delete(path):
    try:
        ret = _vault_cmd(["delete", "-format=json", path])
    except RuntimeError as err:
        pytest.fail(f"Failed to delete path at `{path}`: {err}")
    try:
        return json.loads(ret.stdout) or True
    except json.decoder.JSONDecodeError:
        return True


def vault_list(path):
    ret = _vault_cmd(["list", "-format=json", path], raw=True)
    if ret.returncode != 0:
        if ret.returncode == 2:
            return []
        log.debug("Failed to list path at `%s`:\n%s\nSTDERR: %s", path, ret, ret.stderr)
        pytest.fail(f"Failed to list path at `{path}`: {ret.stderr or ret.stdout}")
    return json.loads(ret.stdout)


def vault_list_detailed(path):
    ret = _vault_cmd(["list", "-detailed", "-format=json", path], raw=True)
    if ret.returncode != 0:
        if ret.returncode == 2:
            return []
        log.debug("Failed to list path at `%s`:\n%s\nSTDERR: %s", path, ret, ret.stderr)
        pytest.fail(f"Failed to list path at `{path}`: {ret.stderr or ret.stdout}")
    return json.loads(ret.stdout)["data"]


def vault_read(path, default=..., raise_errors=False):
    try:
        ret = _vault_cmd(["read", "-format=json", path])
    except RuntimeError as err:
        if raise_errors:
            raise
        if default is not ...:
            return default
        pytest.fail(f"Failed to read path at `{path}`: {err}")
    return json.loads(ret.stdout)


def vault_write(path, *args, **kwargs):
    kwargs_ = [f"{k}={v}" for k, v in kwargs.items()]
    cmd = (
        ["write", "-format=json"]
        + (["-f"] if not (args or kwargs) else [])
        + [path]
        + list(args)
        + kwargs_
    )
    try:
        ret = _vault_cmd(cmd)
    except RuntimeError as err:
        pytest.fail(f"Failed to write to path at `{path}`: {err}")
    try:
        return json.loads(ret.stdout) or True
    except json.decoder.JSONDecodeError:
        return True


def vault_revoke(lease_id, prefix=False):
    cmd = ["lease", "revoke"]
    if prefix:
        cmd += ["-prefix"]
    cmd += [lease_id]
    try:
        _vault_cmd(cmd)
    except RuntimeError as err:
        pytest.fail(f"Failed to revoke lease `{lease_id}`: {err}")
    return True



import logging

import pytest
from salt.exceptions import CommandExecutionError

from tests.support.vault import vault_delete_policy
from tests.support.vault import vault_delete_secret
from tests.support.vault import vault_list_policies
from tests.support.vault import vault_list_secrets
from tests.support.vault import vault_read_policy
from tests.support.vault import vault_read_secret
from tests.support.vault import vault_write_policy
from tests.support.vault import vault_write_secret

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
]

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def minion_config_overrides(vault_port):
    return {
        "vault": {
            "auth": {
                "method": "token",
                "token": "testsecret",
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        }
    }


@pytest.fixture(scope="module")
def sys_mod(modules):
    return modules.sys


@pytest.fixture
def vault(modules, container):  # pylint: disable=unused-argument
    try:
        yield modules.vault
    finally:
        # We're explicitly using the vault CLI and not the salt vault module
        for mount in ("secret", "secret-v1"):
            secret_path = f"{mount}/my"
            for secret in vault_list_secrets(secret_path):
                vault_delete_secret(f"{secret_path}/{secret}", metadata=True)
        policies = vault_list_policies()
        for policy in ("functional_test_policy", "policy_write_test"):
            if policy in policies:
                vault_delete_policy(policy)


def test_vault_read_secret_issue_61084(sys_mod):
    """
    Test issue 61084. `sys.argspec` should return valid data and not throw a
    TypeError due to pickling
    This should probably be a pre-commit check or something
    """
    result = sys_mod.argspec("vault.read_secret")
    assert isinstance(result, dict)
    assert isinstance(result.get("vault.read_secret"), dict)


def test_vault_list_secrets_issue_61084(sys_mod):
    """
    Test issue 61084. `sys.argspec` should return valid data and not throw a
    TypeError due to pickling
    This should probably be a pre-commit check or something
    """
    result = sys_mod.argspec("vault.list_secrets")
    assert isinstance(result, dict)
    assert isinstance(result.get("vault.list_secrets"), dict)


@pytest.fixture
def existing_secret_v1(container):  # pylint: disable=unused-argument
    path = "secret-v1/my/secret"
    vault_write_secret(path, user="foo", password="bar")
    ret = vault_read_secret(path)
    assert ret == {"user": "foo", "password": "bar"}
    return path


@pytest.mark.usefixtures("existing_secret_v1")
def test_read_secret_v1(vault):
    res = vault.read_secret("secret-v1/my/secret")
    assert res == {"user": "foo", "password": "bar"}


def test_write_secret_v1(vault):
    data = {"user": "foo", "password": "bar"}
    assert vault.write_secret("secret-v1/my/written_secret", **data) is True
    assert vault_read_secret("secret-v1/my/written_secret") == data


@pytest.mark.usefixtures("existing_secret_v1")
def test_list_secrets_v1(vault):
    ret = vault.list_secrets("secret-v1/my/")
    assert ret
    assert "keys" in ret
    assert ret["keys"] == ["secret"]


def test_delete_secret_v1(vault, existing_secret_v1):
    assert vault.delete_secret(existing_secret_v1) is True
    assert not vault_list_secrets("/".join(existing_secret_v1.split("/")[:-1]))


def test_write_read_secret(vault):
    ret = vault.write_secret(path="secret/my/secret", user="foo", password="bar")
    # write_secret output:
    # {'created_time': '2020-01-12T23:09:34.571294241Z', 'destroyed': False,
    # 'version': 1, 'deletion_time': ''}
    assert ret
    expected_write = {"destroyed": False, "deletion_time": ""}
    for key in list(ret):
        if key not in expected_write:
            ret.pop(key)
    assert ret == expected_write

    ret = vault.read_secret("secret/my/secret", metadata=True)
    # read_secret output:
    # {'data': {'password': 'bar', 'user': 'foo'},
    # 'metadata': {'created_time': '2020-01-12T23:07:18.829326918Z', 'destroyed': False,
    # 'version': 1, 'deletion_time': ''}}
    assert ret
    assert "data" in ret
    expected_read = {"password": "bar", "user": "foo"}
    assert "metadata" in ret
    assert ret["data"] == expected_read

    ret = vault.read_secret("secret/my/secret")
    for key in list(ret):
        if key not in expected_read:
            ret.pop(key)
    assert ret == expected_read

    ret = vault.read_secret("secret/my/secret", key="user")
    assert ret == "foo"


def test_write_raw_read_secret(vault):
    ret = vault.write_raw("secret/my/secret2", raw={"user2": "foo2", "password2": "bar2"})
    assert ret
    expected_write = {"destroyed": False, "deletion_time": ""}
    for key in list(ret):
        if key not in expected_write:
            ret.pop(key)
    assert ret == expected_write

    expected_read = {"password2": "bar2", "user2": "foo2"}
    ret = vault.read_secret("secret/my/secret2", metadata=True)
    assert ret
    assert "metadata" in ret
    assert "data" in ret
    assert ret["data"] == expected_read

    ret = vault.read_secret("secret/my/secret2")
    for key in list(ret):
        if key not in expected_read:
            ret.pop(key)
    assert ret == expected_read


@pytest.fixture
def existing_secret(vault):
    secret_key = "secret/my/secret"
    ret = vault.write_secret(secret_key, user="foo", password="bar")
    expected_write = {"destroyed": False, "deletion_time": ""}
    for key in list(ret):
        if key not in expected_write:
            ret.pop(key)
    assert ret == expected_write
    try:
        yield
    finally:
        # ensure we don't retain the version history between tests
        vault_delete_secret(secret_key, metadata=True)


@pytest.fixture
def existing_secret_version(existing_secret, vault):  # pylint: disable=unused-argument
    ret = vault.write_secret("secret/my/secret", user="foo", password="hunter1")
    assert ret
    assert ret["version"] == 2
    ret = vault.read_secret("secret/my/secret")
    assert ret
    assert ret["password"] == "hunter1"


@pytest.mark.usefixtures("existing_secret_version")
def test_read_secret_version(vault):
    ret = vault.read_secret("secret/my/secret", version=1)
    assert ret["password"] == "bar"
    ret = vault.read_secret("secret/my/secret", version=2)
    assert ret["password"] == "hunter1"


@pytest.mark.usefixtures("existing_secret")
def test_delete_restore_secret(vault):
    ret = vault.delete_secret("secret/my/secret")
    assert ret is True
    ret = vault.read_secret("secret/my/secret", default="__was_deleted__")
    assert ret == "__was_deleted__"
    ret = vault.restore_secret("secret/my/secret")
    assert ret is True
    ret = vault.read_secret("secret/my/secret")
    assert ret["password"] == "bar"


@pytest.mark.usefixtures("existing_secret_version")
def test_delete_restore_secret_versions(vault):
    key = "secret/my/secret"
    # delete the first version only, the current one is version 2
    assert vault.delete_secret(key, 1) is True
    ret = vault.read_secret(key)
    assert ret
    assert ret["password"] == "hunter1"
    ret = vault.read_secret(key, default="__was_deleted__", version=1)
    assert ret == "__was_deleted__"
    # now delete the current one as well
    assert vault.delete_secret(key, 2) is True
    ret = vault.read_secret(key, default="__was_deleted__")
    assert ret == "__was_deleted__"
    assert vault.restore_secret(key, 1) is True
    ret = vault.read_secret(key, version=1)
    assert ret["password"] == "bar"
    assert vault.restore_secret(key, 2) is True
    ret = vault.read_secret(key)
    assert ret["password"] == "hunter1"


@pytest.mark.usefixtures("existing_secret_version")
def test_delete_restore_secret_all_versions(vault):
    versions = ("bar", "hunter1")
    assert vault.delete_secret("secret/my/secret", all_versions=True) is True
    for version in range(len(versions)):
        ret = vault.read_secret("secret/my/secret", default="__was_deleted__", version=version + 1)
        assert ret == "__was_deleted__"
    assert vault.restore_secret("secret/my/secret", all_versions=True) is True
    for version, data in enumerate(versions):
        ret = vault.read_secret("secret/my/secret", version=version + 1)
        assert ret["password"] == data


@pytest.mark.usefixtures("existing_secret_version")
def test_restore_secret_latest_not_deleted(vault):
    assert vault.delete_secret("secret/my/secret", 1) is True
    with pytest.raises(CommandExecutionError, match="No secret version to restore."):
        vault.restore_secret("secret/my/secret")


@pytest.mark.usefixtures("existing_secret")
def test_list_secrets(vault):
    ret = vault.list_secrets("secret/my/")
    assert ret
    assert "keys" in ret
    assert ret["keys"] == ["secret"]


@pytest.mark.usefixtures("existing_secret_version")
def test_destroy_secret(vault):
    assert vault.destroy_secret("secret/my/secret") is True
    ret = vault.read_secret("secret/my/secret", default="__was_deleted__")
    assert ret == "__was_deleted__"
    ret = vault.read_secret("secret/my/secret", version=1)
    assert ret["password"] == "bar"


@pytest.mark.usefixtures("existing_secret_version")
def test_destroy_secret_versions(vault):
    assert vault.destroy_secret("secret/my/secret", "1") is True
    ret = vault.read_secret("secret/my/secret")
    assert ret["password"] == "hunter1"


@pytest.mark.usefixtures("existing_secret_version")
def test_destroy_secret_all_versions(vault):
    assert vault.destroy_secret("secret/my/secret", all_versions=True) is True
    for version in range(2):
        ret = vault.read_secret("secret/my/secret", default="__was_deleted__", version=version + 1)
        assert ret == "__was_deleted__"


@pytest.mark.usefixtures("existing_secret_version")
def test_wipe_secret(vault):
    assert vault.wipe_secret("secret/my/secret") is True
    assert vault.read_secret_meta("secret/my/secret") is False


@pytest.mark.usefixtures("existing_secret")
def test_patch_secret(vault):
    ret = vault.patch_secret("secret/my/secret", password="baz")
    assert ret
    expected_write = {"destroyed": False, "deletion_time": ""}
    for key in list(ret):
        if key not in expected_write:
            ret.pop(key)
    assert ret == expected_write
    ret = vault.read_secret("secret/my/secret")
    assert ret == {"user": "foo", "password": "baz"}


@pytest.fixture
def policy_rules():
    return """\
path "secret/some/thing" {
    capabilities = ["read"]
}
    """


@pytest.fixture
def existing_policy(policy_rules, container):  # pylint: disable=unused-argument
    vault_write_policy("functional_test_policy", policy_rules)
    try:
        yield
    finally:
        vault_delete_policy("functional_test_policy")


@pytest.mark.usefixtures("existing_policy")
def test_policy_fetch(vault, policy_rules):
    ret = vault.policy_fetch("functional_test_policy")
    assert ret == policy_rules
    ret = vault.policy_fetch("__does_not_exist__")
    assert ret is None


def test_policy_write(vault, policy_rules):
    ret = vault.policy_write("policy_write_test", policy_rules)
    assert ret is True
    assert vault_read_policy("policy_write_test") == policy_rules


@pytest.mark.usefixtures("existing_policy")
def test_policy_delete(vault):
    ret = vault.policy_delete("functional_test_policy")
    assert ret is True
    assert "functional_test_policy" not in vault_list_policies()


@pytest.mark.usefixtures("existing_policy")
def test_policies_list(vault):
    ret = vault.policies_list()
    assert "functional_test_policy" in ret



"""
Beacon for the Vault integration. Sends events when a
lease's TTL undercuts a specified value. By default, also
tries to renew leases before sending an event.

.. versionadded:: 1.1.0

Event description
-----------------

When a lease undercuts its minimum TTL, an event is sent.

The event tag's format is: ``salt/beacon/<minion ID>/vault_lease_<lease cache key>/expire``

The event data contains (non-exhaustive):

* ``expires_in`` - number of seconds left until the lease is revoked by Vault (can be ``-1`` if already revoked)
* ``lease_id`` - the lease ID of the expiring lease
* ``ckey`` - the cache key of the expiring lease
* ``meta`` - custom metadata, e.g. for use in a reactor
* ``expired`` - if the lease is already expired

Example configuration
---------------------
.. code-block:: yaml

    beacons:
      vault_basic:
        - beacon_module: vault_lease
        - leases:
            - db.database.dynamic.basic_lease.default

      vault_advanced:
        - beacon_module: vault_lease
        - leases:
            db.database.dynamic.write_stuff.default: {}
            db.database.dynamic.monitoring.default:
              renew: false
            db.database.dynamic.read_stuff.default:
              min_ttl: 6h
              meta:
                sls: read.stuff
        - min_ttl: 1h
        - meta:
            sls: write.stuff
        - check_server: true

.. _beacon-state-example:

Example for enabling beacon via state
-------------------------------------
This beacon can be added dynamically when explicitly caching
database leases.

.. code-block:: yaml

    Important Vault lease is cached:
      vault_db.creds_cached:
        - name: my_important_role
        - valid_for: 6h  # minimum TTL for the lease to be returned by get_creds
        - revoke_delay: 30m
        - beacon: true   # also add a beacon for monitoring
        - beacon_interval: 300  # interval between beacon runs
        - min_ttl: 12h   # minimum TTL for the beacon to accept the lease as valid
        - meta: my.important.state  # can be used with a reactor
        - order: first   # leases should be cached early

Configuration reference
-----------------------
.. vconf:: lease_beacon.leases

``leases``
    The leases to monitor, referenced by their cache keys.
    This can be a string (single lease), list (multiple leases)
    or mapping (multiple leases with parameter overrides).

.. vconf:: lease_beacon.min_ttl

``min_ttl``
    The minimum TTL a monitored lease should have.
    Can be overridden per configured lease in :vconf:`lease_beacon.leases`.
    If a ``min_ttl`` was set on the lease during its creation,
    this value must be equal or greater to have any effect.
    Defaults to ``300``.

.. vconf:: lease_beacon.check_server

``check_server``
    Whether cached leases should be validated with the Vault server
    before declaring them as valid.
    Can be overridden per configured lease in :vconf:`lease_beacon.leases`.
    There is no equivalent parameter that can be set on the lease during
    its creation currently.
    Defaults to false.

.. vconf:: lease_beacon.meta

``meta``
    Arbitrary metadata to include in expiry events.
    Can be overridden per configured lease in :vconf:`lease_beacon.leases`.
    If ``meta`` was set on the lease during creation, the corresponding
    value takes precedence. If both values are either mappings or lists,
    they will be merged together.

.. vconf:: lease_beacon.renew

``renew``
    Before sending an event, try to renew the lease as needed.
    Defaults to true.
"""

import logging

import salt.utils.beacons
import salt.utils.dictupdate as dup

from saltext.vault.utils import vault
from saltext.vault.utils.vault.helpers import timestring_map

log = logging.getLogger(__name__)


__virtualname__ = "vault_lease"


def __virtual__():
    return __virtualname__


def validate(config):
    """
    Validate the beacon configuration
    """
    if not isinstance(config, list):
        return False, "Configuration for vault_lease must be a list"
    config = salt.utils.beacons.list_to_dict(config)
    if "leases" not in config:
        return False, "Requires monitored lease(s) cache key(s) in `leases`"
    if not isinstance(config["leases"], (dict, list, str)):
        return False, "`leases` must be a dict, list or str"

    if isinstance(config["leases"], str):
        if "*" in config["leases"]:
            return False, "`leases` does not support globs"
    else:
        if any("*" in lease for lease in config["leases"]):
            return False, "`leases` does not support globs"
        if isinstance(config["leases"], dict) and any(
            not isinstance(cfg, dict) for cfg in config["leases"].values()
        ):
            return False, "`leases` mapping values must be dicts"

    return True, "Valid beacon configuration."


def beacon(config):
    """
    Watch the configured lease(s).
    """
    config = _render_config(config)
    # background processes should not pass __context__
    store = vault.get_lease_store(__opts__, {})
    events = []
    for lease, lease_config in config["leases"].items():
        info = store.list_info(match=lease)
        if not info:
            events.append(_enrich_info(lease, lease_config, {"expires_in": -1, "expired": True}))
            continue
        lease_info = info[lease]
        effective_config = _merge_lease_config(lease_config, lease_info)
        if effective_config.get("check_server"):
            try:
                store.lookup(lease_info["lease_id"])
            except vault.VaultNotFoundError:
                store.revoke(lease_info["lease_id"], delta=lease_info.get("revoke_delay"))
                lease_info["expires_in"] = -1
                lease_info["expired"] = True
                events.append(_enrich_info(lease, effective_config, lease_info))
                continue
        if lease_info["expired"]:
            events.append(_enrich_info(lease, effective_config, lease_info))
            continue
        if timestring_map(effective_config["min_ttl"]) >= lease_info["expires_in"]:
            if not effective_config.get("renew", True):
                events.append(_enrich_info(lease, effective_config, lease_info))
                continue
            # attempt renewal
            res = store.get(
                lease,
                valid_for=effective_config["min_ttl"],
                revoke=False,
                check_server=effective_config.get("check_server", False),
            )
            if not res:
                events.append(_enrich_info(lease, effective_config, lease_info))
                continue
    return events


def _enrich_info(lease, effective_config, info):
    info["ckey"] = lease
    info["meta"] = effective_config.get("meta")
    info["min_ttl"] = effective_config.get("min_ttl", 300)
    info["check_server"] = effective_config.get("check_server")
    info.pop("id", None)
    info["tag"] = "expire"
    return info


def _render_config(cfg):
    config = salt.utils.beacons.list_to_dict(cfg)
    if isinstance(config["leases"], str):
        config["leases"] = {config["leases"]: {}}
    if not isinstance(config["leases"], dict):
        config["leases"] = {lease: {} for lease in config["leases"]}
    defaults = {}
    for param in ("min_ttl", "meta", "check_server", "renew"):
        if param in config:
            defaults[param] = config[param]
    return {
        "leases": {
            lease: {**defaults, **lease_config} for lease, lease_config in config["leases"].items()
        }
    }


def _merge_lease_config(cfg, lease):
    if cfg.get("min_ttl") is not None and lease.get("min_ttl") is not None:
        cfg["min_ttl"] = (
            lease["min_ttl"]
            if timestring_map(lease["min_ttl"]) >= timestring_map(cfg["min_ttl"])
            else cfg["min_ttl"]
        )
    elif lease.get("min_ttl") is not None:
        cfg["min_ttl"] = lease["min_ttl"]
    elif "min_ttl" not in cfg:
        cfg["min_ttl"] = 300
    cfg["meta"] = _merge_meta(cfg.get("meta"), lease.get("meta"))
    return cfg


def _merge_meta(default, ovrr):
    if ovrr is None:
        return default
    default = default if default is not None else {}
    for val in (default, ovrr):
        if not isinstance(val, (dict, list)):
            return ovrr
    if type(default) is not type(ovrr):
        return ovrr
    if isinstance(default, list):
        return default + ovrr
    return dup.merge(default, ovrr, merge_lists=True) or None



import pytest

from tests.support.vault import vault_delete
from tests.support.vault import vault_disable_secret_engine
from tests.support.vault import vault_enable_secret_engine
from tests.support.vault import vault_list
from tests.support.vault import vault_read
from tests.support.vault import vault_write

try:
    from cryptography.hazmat.primitives.serialization import SSHCertificateType
    from cryptography.hazmat.primitives.serialization import load_ssh_public_identity

    CERT_CHECK = True
except ImportError:
    CERT_CHECK = False

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def master_config_overrides():
    return {
        "vault": {
            "policies": {
                "assign": [
                    "salt_minion",
                    "ssh_admin",
                ]
            },
        },
    }


@pytest.fixture
def userrole():
    return {
        "key_type": "ca",
        "allowed_users": "*",
        "allowed_extensions": "*",
        "allow_user_certificates": True,
        "ttl": 3600,
        "max_ttl": 86400,
    }


@pytest.fixture
def hostrole():
    return {
        "key_type": "ca",
        "allowed_domains": "*",
        "allow_host_certificates": True,
        "ttl": 3600,
        "max_ttl": 86400,
    }


@pytest.fixture
def iprole():
    return {
        "key_type": "otp",
        "default_user": "foobar",
        "cidr_list": "0.0.0.0/1",
        "exclude_cidr_list": "128.0.0.0/1",
        "port": 9876,
    }


@pytest.fixture
def ec_priv():
    return """
-----BEGIN OPENSSH PRIVATE KEY-----
b3BlbnNzaC1rZXktdjEAAAAABG5vbmUAAAAEbm9uZQAAAAAAAAABAAAAaAAAABNlY2RzYS1zaGEy
LW5pc3RwMjU2AAAACG5pc3RwMjU2AAAAQQSO2hM3nJP6fxgzyXIEEKhHeqOqXccIvV8EZLfqCrcX
NGR7Be7yYdPNx+bcNFx5fyLrxKin+f/4pV4/q+mMoqVxAAAAoHwE+2V8BPtlAAAAE2VjZHNhLXNo
YTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBI7aEzeck/p/GDPJcgQQqEd6o6pdxwi9XwRkt+oK
txc0ZHsF7vJh083H5tw0XHl/IuvEqKf5//ilXj+r6YyipXEAAAAgc5sbtq4PEIHmkqJzEbNaO1sB
2PSfCjWqlGSC7ODBqy4AAAAAAQIDBAUGBwg=
-----END OPENSSH PRIVATE KEY-----
    """.strip()


@pytest.fixture
def ec_priv_file(ec_priv, tmp_path):
    path = tmp_path / "ec"
    path.write_text(ec_priv)
    return str(path)


@pytest.fixture
def ec_pub():
    return "ecdsa-sha2-nistp256 AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBI7aEzeck/p/GDPJcgQQqEd6o6pdxwi9XwRkt+oKtxc0ZHsF7vJh083H5tw0XHl/IuvEqKf5//ilXj+r6YyipXE="


@pytest.fixture
def ec_pub_file(ec_pub, tmp_path):
    path = tmp_path / "ec.pub"
    path.write_text(ec_pub)
    return str(path)


@pytest.fixture(scope="module", autouse=True)
def ssh_engine(container):  # pylint: disable=unused-argument
    assert vault_enable_secret_engine("ssh")
    yield
    assert vault_disable_secret_engine("ssh")


@pytest.fixture(params=(("userrole", "hostrole", "iprole"),))
def roles_setup(request):  # pylint: disable=unused-argument
    try:
        for role_name in request.param:
            role_args = request.getfixturevalue(role_name)
            vault_write(f"ssh/roles/{role_name}", **role_args)
            assert role_name in vault_list("ssh/roles")
        yield
    finally:
        for role_name in request.param:
            if role_name in vault_list("ssh/roles"):
                vault_delete(f"ssh/roles/{role_name}")
                assert role_name not in vault_list("ssh/roles")


@pytest.mark.usefixtures("roles_setup")
def test_read_role(salt_ssh_cli):
    res = salt_ssh_cli.run("vault_ssh.read_role", "userrole")
    assert res.returncode == 0
    expected = {
        "algorithm_signer": "default",
        "allow_bare_domains": False,
        "allow_host_certificates": False,
        "allow_subdomains": False,
        "allow_user_certificates": True,
        "allow_user_key_ids": False,
        "allowed_critical_options": "",
        "allowed_domains": "",
        "allowed_domains_template": False,
        "allowed_extensions": "*",
        "allowed_user_key_lengths": {},
        "allowed_users": "*",
        "allowed_users_template": False,
        "default_critical_options": {},
        "default_extensions": {},
        "default_extensions_template": False,
        "default_user": "",
        "default_user_template": False,
        "key_id_format": "",
        "key_type": "ca",
        "max_ttl": 86400,
        "ttl": 3600,
    }
    for var, val in expected.items():
        assert var in res.data
        assert res.data[var] == val


@pytest.fixture
def _temp_role():
    name = "testrole"
    try:
        yield name
    finally:
        vault_delete(f"ssh/roles/{name}")


def test_write_role_ca(salt_ssh_cli, userrole, _temp_role):
    key_type = userrole.pop("key_type")
    res = salt_ssh_cli.run("vault_ssh.write_role_ca", _temp_role, **userrole)
    assert res.returncode == 0
    assert res.data is True
    data = vault_read(f"ssh/roles/{_temp_role}")["data"]
    assert data["key_type"] == key_type
    for var, val in userrole.items():
        assert var in data
        assert data[var] == val


def test_write_role_otp(salt_ssh_cli, iprole, _temp_role):
    key_type = iprole.pop("key_type")
    res = salt_ssh_cli.run("vault_ssh.write_role_otp", _temp_role, **iprole)
    assert res.returncode == 0
    assert res.data is True
    data = vault_read(f"ssh/roles/{_temp_role}")["data"]
    assert data["key_type"] == key_type
    for var, val in iprole.items():
        assert var in data
        assert data[var] == val


@pytest.mark.usefixtures("roles_setup")
def test_delete_role(salt_ssh_cli):
    res = salt_ssh_cli.run("vault_ssh.delete_role", "userrole")
    assert res.returncode == 0
    assert res.data is True
    assert "userrole" not in vault_list("ssh/roles")


@pytest.mark.usefixtures("roles_setup")
def test_list_roles(salt_ssh_cli):
    res = salt_ssh_cli.run("vault_ssh.list_roles")
    assert res.returncode == 0
    for role in ("userrole", "hostrole", "iprole"):
        assert role in res.data


@pytest.mark.usefixtures("roles_setup")
def test_list_roles_ip(salt_ssh_cli):
    res = salt_ssh_cli.run("vault_ssh.list_roles_ip", "10.1.0.1")
    assert res.returncode == 0
    assert res.data == ["iprole"]


@pytest.mark.usefixtures("roles_setup")
def test_zeroaddress_roles(salt_ssh_cli):
    res = salt_ssh_cli.run("vault_ssh.list_roles_zeroaddr")
    assert res.returncode == 0
    assert res.data == []
    res = salt_ssh_cli.run("vault_ssh.write_zeroaddr_roles", ["iprole"])
    assert res.returncode == 0
    assert res.data is True
    res = salt_ssh_cli.run("vault_ssh.list_roles_zeroaddr")
    assert res.returncode == 0
    assert res.data == ["iprole"]
    res = salt_ssh_cli.run("vault_ssh.delete_zeroaddr_roles")
    assert res.returncode == 0
    assert res.data is True
    res = salt_ssh_cli.run("vault_ssh.list_roles_zeroaddr")
    assert res.returncode == 0
    assert res.data == []


@pytest.fixture
def _temp_ca():
    try:
        yield
    finally:
        vault_delete("ssh/config/ca")


@pytest.mark.usefixtures("_temp_ca")
def test_create_ca(salt_ssh_cli):
    res = salt_ssh_cli.run("vault_ssh.create_ca")
    assert res.returncode == 0
    assert res.data.startswith("ssh-rsa ")
    assert "public_key" in vault_read("ssh/config/ca")["data"]


@pytest.mark.usefixtures("_temp_ca")
def test_create_ca_key_spec(salt_ssh_cli):
    res = salt_ssh_cli.run("vault_ssh.create_ca", key_type="ec", key_bits=384)
    assert res.returncode == 0
    assert res.data.startswith("ecdsa-")
    assert "p384" in res.data
    assert "public_key" in vault_read("ssh/config/ca")["data"]


@pytest.mark.usefixtures("_temp_ca")
def test_create_ca_with_keys(salt_ssh_cli, ec_pub, ec_priv_file):
    res = salt_ssh_cli.run("vault_ssh.create_ca", public_key=ec_pub, private_key=ec_priv_file)
    assert res.returncode == 0
    assert res.data == ec_pub
    assert "public_key" in vault_read("ssh/config/ca")["data"]


@pytest.fixture
def ca_setup(ec_priv, ec_pub):
    vault_write("ssh/config/ca", private_key=ec_priv, public_key=ec_pub)
    assert vault_read("ssh/config/ca", default=False)
    try:
        yield
    finally:
        vault_delete("ssh/config/ca")


@pytest.mark.usefixtures("ca_setup")
def test_read_ca(salt_ssh_cli, ec_pub):
    res = salt_ssh_cli.run("vault_ssh.read_ca")
    assert res.returncode == 0
    assert res.data == ec_pub


@pytest.mark.usefixtures("ca_setup")
def test_destroy_ca(salt_ssh_cli, container):
    res = salt_ssh_cli.run("vault_ssh.destroy_ca")
    assert res.returncode == 0
    if "openbao" in container:
        assert res.data["warnings"]
        assert "Deleted 1 issuers" in res.data["warnings"][0]
    else:
        assert res.data is True
    try:
        res = vault_read("ssh/config/ca", raise_errors=True)
    except RuntimeError as err:
        if "openbao" in container:
            assert "no default issuer currently configured" in str(err)
        else:
            assert "keys haven't been configured yet" in str(err)
    else:
        raise AssertionError(f"CA has not been deleted: {res}")


@pytest.mark.usefixtures("ca_setup", "roles_setup")
def test_sign_key_user(salt_ssh_cli, ec_pub, container):
    res = salt_ssh_cli.run(
        "vault_ssh.sign_key",
        "userrole",
        ec_pub,
        critical_options={"force-command": "rm -rf /"},
        extensions={"permit-pty": ""},
        valid_principals=["foobar"],
    )
    assert res.returncode == 0
    expected = {"serial_number", "signed_key"}
    if "openbao" in container:
        expected.add("issuer_id")
    assert set(res.data) == expected
    if CERT_CHECK:
        cert = load_cert(res.data["signed_key"])
        assert cert.type == SSHCertificateType.USER
        assert cert.critical_options == {b"force-command": b"rm -rf /"}
        assert cert.extensions == {b"permit-pty": b""}
        assert cert.valid_principals == [b"foobar"]


@pytest.mark.usefixtures("ca_setup", "roles_setup")
def test_sign_key_host(salt_ssh_cli, ec_pub, container):
    res = salt_ssh_cli.run(
        "vault_ssh.sign_key", "hostrole", ec_pub, cert_type="host", valid_principals=["foo.bar.biz"]
    )
    assert res.returncode == 0
    expected = {"serial_number", "signed_key"}
    if "openbao" in container:
        expected.add("issuer_id")
    assert set(res.data) == expected
    if CERT_CHECK:
        cert = load_cert(res.data["signed_key"])
        assert cert.type == SSHCertificateType.HOST
        assert not cert.critical_options
        assert cert.valid_principals == [b"foo.bar.biz"]


@pytest.mark.usefixtures("ca_setup", "roles_setup")
def test_generate_key_cert_user(salt_ssh_cli, container):
    res = salt_ssh_cli.run(
        "vault_ssh.generate_key_cert",
        "userrole",
        critical_options={"force-command": "rm -rf /"},
        extensions={"permit-pty": ""},
        valid_principals=["foobar"],
    )
    assert res.returncode == 0
    expected = {"private_key", "private_key_type", "serial_number", "signed_key"}
    if "openbao" in container:
        expected.add("issuer_id")
    assert set(res.data) == expected
    assert res.data["private_key_type"] == "ssh-rsa"
    assert res.data["private_key"].startswith("-----BEGIN")
    if CERT_CHECK:
        cert = load_cert(res.data["signed_key"])
        assert cert.type == SSHCertificateType.USER
        assert cert.critical_options == {b"force-command": b"rm -rf /"}
        assert cert.extensions == {b"permit-pty": b""}
        assert cert.valid_principals == [b"foobar"]


@pytest.mark.usefixtures("ca_setup", "roles_setup")
def test_generate_key_cert_host(salt_ssh_cli, container):
    res = salt_ssh_cli.run(
        "vault_ssh.generate_key_cert",
        "hostrole",
        cert_type="host",
        valid_principals=["foo.bar.biz"],
    )
    assert res.returncode == 0
    expected = {"private_key", "private_key_type", "serial_number", "signed_key"}
    if "openbao" in container:
        expected.add("issuer_id")
    assert set(res.data) == expected
    assert res.data["private_key_type"] == "ssh-rsa"
    assert res.data["private_key"].startswith("-----BEGIN")
    if CERT_CHECK:
        cert = load_cert(res.data["signed_key"])
        assert cert.type == SSHCertificateType.HOST
        assert not cert.critical_options
        assert cert.valid_principals == [b"foo.bar.biz"]


def load_cert(data):
    return load_ssh_public_identity(data.encode())



"""
Very simple heuristic to generate the next version number
based on the current changelog news fragments.

This looks for the most recent version by parsing the
CHANGELOG.md file and increments a specific part,
depending on the fragment types present and their contents.

Major bumps are caused by:
  * files named `.removed.md`
  * files named `.breaking.md`
  * files containing `BREAKING:`

Minor bumps are caused by:
  * files named `.added.md`

Otherwise, only the patch version is bumped.
"""

import re
import sys
from pathlib import Path

PROJECT_ROOT = Path(".").resolve()
CHANGELOG_DIR = PROJECT_ROOT / "changelog"
CHANGELOG_FILE = PROJECT_ROOT / "CHANGELOG.md"


class Version:
    def __init__(self, version):
        match = re.search(r"v?(?P<release>[0-9]+(?:\.[0-9]+)*)", version)
        if not match:
            raise ValueError(f"Invalid version: '{version}'")
        self.release = tuple(int(i) for i in match.group("release").split("."))

    @property
    def major(self):
        return self._ret(0)

    @property
    def minor(self):
        return self._ret(1)

    @property
    def patch(self):
        return self._ret(2)

    def __str__(self):
        return ".".join(str(i) for i in self.release)

    def _ret(self, cnt):
        try:
            return self.release[cnt]
        except IndexError:
            return 0


def last_release():
    for line in CHANGELOG_FILE.read_text(encoding="utf-8").splitlines():
        if line.startswith("## "):
            return Version(line.split(" ")[1])
    return Version("0.0.0")


def get_next_version(last):
    major = minor = False

    for fragment in CHANGELOG_DIR.glob("[!.]*"):
        name = fragment.name.lower()
        if ".added" in name:
            minor = True
        elif ".breaking" in name or ".removed" in name:
            major = True
            break
        if "breaking:" in fragment.read_text(encoding="utf-8").lower():
            major = True
            break
    if major:
        return Version(f"{last.major + 1}.0.0")
    if minor:
        return Version(f"{last.major}.{last.minor + 1}.0")
    return Version(f"{last.major}.{last.minor}.{last.patch + 1}")


if __name__ == "__main__":
    try:
        if sys.argv[1] == "next":
            print(get_next_version(last_release()))
            raise SystemExit(0)
    except IndexError:
        pass
    print(last_release())



import contextlib
import copy
from unittest.mock import ANY
from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import call
from unittest.mock import patch

import pytest
import requests
import salt.exceptions
import urllib3.exceptions

from saltext.vault.utils import vault
from saltext.vault.utils.vault import client as vclient
from tests.unit.utils.vault.conftest import _mock_json_response  # pylint: disable=import-error


@pytest.mark.parametrize(
    "endpoint",
    [
        "secret/some/path",
        "/secret/some/path",
        "secret/some/path/",
        "/secret/some/path/",
    ],
)
def test_vault_client_request_raw_url(endpoint, client, req):
    """
    Test that requests are sent to the correct endpoint, regardless of leading or trailing slashes
    """
    expected_url = f"{client.url}/v1/secret/some/path"
    client.request_raw("GET", endpoint)
    req.assert_called_with(
        "GET",
        expected_url,
        headers=ANY,
        json=None,
    )


def test_vault_client_request_raw_kwargs_passthrough(client, req):
    """
    Test that kwargs for requests.request are passed through
    """
    client.request_raw(
        "GET", "secret/some/path", allow_redirects=False, cert="/etc/certs/client.pem"
    )
    req.assert_called_with(
        "GET",
        ANY,
        headers=ANY,
        json=ANY,
        allow_redirects=False,
        cert="/etc/certs/client.pem",
    )


@pytest.mark.parametrize("namespace", [None, "test-namespace"])
@pytest.mark.parametrize("client", [None], indirect=True)
def test_vault_client_request_raw_headers_namespace(namespace, client, req):
    """
    Test that namespace is present in the HTTP headers only if it was specified
    """
    if namespace is not None:
        client.namespace = namespace

    namespace_header = "X-Vault-Namespace"
    client.request_raw("GET", "secret/some/path")
    headers = req.call_args.kwargs.get("headers", {})
    if namespace is None:
        assert namespace_header not in headers
    else:
        assert headers.get(namespace_header) == namespace


@pytest.mark.parametrize("wrap", [False, 30, "1h"])
def test_vault_client_request_raw_headers_wrap(wrap, client, req):
    """
    Test that the wrap header is present only if it was specified and supports time strings
    """
    wrap_header = "X-Vault-Wrap-TTL"
    client.request_raw("GET", "secret/some/path", wrap=wrap)
    headers = req.call_args.kwargs.get("headers", {})
    if not wrap:
        assert wrap_header not in headers
    else:
        assert headers.get(wrap_header) == str(wrap)


@pytest.mark.parametrize("header", ["X-Custom-Header", "X-Existing-Header"])
def test_vault_client_request_raw_headers_additional(header, client, req):
    """
    Test that additional headers are passed correctly and override default ones
    """
    with patch.object(
        client, "_get_headers", Mock(return_value={"X-Existing-Header": "unchanged"})
    ):
        client.request_raw("GET", "secret/some/path", add_headers={header: "changed"})
        actual_header = req.call_args.kwargs.get("headers", {}).get(header)
        assert actual_header == "changed"


@pytest.mark.usefixtures("req_failed")
@pytest.mark.parametrize(
    "req_failed",
    [400, 403, 404, 502, 401],
    indirect=True,
)
@pytest.mark.parametrize(
    "client",
    [None],
    indirect=True,
)
def test_vault_client_request_raw_does_not_raise_http_exception(client):
    """
    request_raw should return the raw response object regardless of HTTP status code
    """
    res = client.request_raw("GET", "secret/some/path")
    with pytest.raises(requests.exceptions.HTTPError):
        res.raise_for_status()


@pytest.mark.usefixtures("req_failed")
@pytest.mark.parametrize(
    "req_failed,expected",
    [
        (400, vault.VaultInvocationError),
        (403, vault.VaultPermissionDeniedError),
        (404, vault.VaultNotFoundError),
        (405, vault.VaultUnsupportedOperationError),
        (412, vault.VaultPreconditionFailedError),
        (429, vault.VaultRateLimitExceededError),
        (500, vault.VaultServerError),
        (502, vault.VaultServerError),
        (503, vault.VaultUnavailableError),
        (401, requests.exceptions.HTTPError),
    ],
    indirect=["req_failed"],
)
@pytest.mark.parametrize("raise_error", [True, False])
def test_vault_client_request_respects_raise_error(raise_error, expected, client):
    """
    request should inspect the response object and raise appropriate errors
    or fall back to raise_for_status if raise_error is true
    """
    if raise_error:
        with pytest.raises(expected):
            client.request("GET", "secret/some/path", raise_error=raise_error)
    else:
        res = client.request("GET", "secret/some/path", raise_error=raise_error)
        assert "errors" in res


def test_vault_client_request_returns_whole_response_data(role_id_response, req, client):
    """
    request should return the whole returned payload, not auth/data etc only
    """
    req.return_value = _mock_json_response(role_id_response)
    res = client.request("GET", "auth/approle/role/test-minion/role-id")
    assert res == role_id_response


def test_vault_client_request_hydrates_wrapped_response(wrapped_role_id_response, req, client):
    """
    request should detect wrapped responses and return an instance of VaultWrappedResponse
    instead of raw data
    """
    req.return_value = _mock_json_response(wrapped_role_id_response)
    res = client.request("GET", "auth/approle/role/test-minion/role-id", wrap="180s")
    assert isinstance(res, vault.VaultWrappedResponse)


@pytest.mark.usefixtures("req_success")
def test_vault_client_request_returns_true_when_no_data_is_reported(client):
    """
    HTTP 204 indicates success with no data returned
    """
    res = client.request("GET", "secret/some/path")
    assert res is True


def test_vault_client_get_config(server_config, client):
    """
    The returned configuration should match the one used to create an instance of VaultClient
    """
    assert client.get_config() == server_config


@pytest.mark.parametrize("client", [None], indirect=["client"])
def test_vault_client_token_valid_false(client):
    """
    The unauthenticated client should always report the token as being invalid
    """
    assert client.token_valid() is False


@pytest.mark.parametrize("client", ["valid_token", "invalid_token"], indirect=True)
@pytest.mark.parametrize("req_any", [200, 403], indirect=True)
@pytest.mark.parametrize("remote", [False, True])
def test_vault_client_token_valid(client, remote, req_any):
    valid = client.token_valid(remote=remote)
    if not remote or not client.auth.is_valid():
        req_any.assert_not_called()
    else:
        req_any.assert_called_once()
    should_be_valid = client.auth.is_valid() and (
        not remote or req_any("POST", "abc").status_code == 200
    )
    assert valid is should_be_valid


@pytest.mark.parametrize("func", ["get", "delete", "post", "list"])
def test_vault_client_wrapper_should_not_require_payload(func, client, req):
    """
    Check that wrappers for get/delete/post/list do not require a payload
    """
    req.return_value = _mock_json_response({}, status_code=200)
    tgt = getattr(client, func)
    res = tgt("auth/approle/role/test-role/secret-id")
    assert res == {}


@pytest.mark.parametrize("func", ["patch"])
def test_vault_client_wrapper_should_require_payload(func, client, req):
    """
    Check that patch wrapper does require a payload
    """
    req.return_value = _mock_json_response({}, status_code=200)
    tgt = getattr(client, func)
    with pytest.raises(TypeError):
        tgt("auth/approle/role/test-role/secret-id")


def test_vault_client_wrap_info_only_data(wrapped_role_id_lookup_response, client, req):
    """
    wrap_info should only return the data portion of the returned wrapping information
    """
    req.return_value = _mock_json_response(wrapped_role_id_lookup_response)
    res = client.wrap_info("test-wrapping-token")
    assert res == wrapped_role_id_lookup_response["data"]


@pytest.mark.usefixtures("req_failed")
@pytest.mark.parametrize(
    "req_failed,expected", [(502, vault.VaultServerError)], indirect=["req_failed"]
)
def test_vault_client_wrap_info_should_fail_with_sensible_response(expected, client):
    """
    wrap_info should return sensible Exceptions, not KeyError etc
    """
    with pytest.raises(expected):
        client.wrap_info("test-wrapping-token")


def test_vault_client_unwrap_returns_whole_response(role_id_response, client, req):
    """
    The unwrapped response should be returned as a whole, not auth/data etc only
    """
    req.return_value = _mock_json_response(role_id_response)
    res = client.unwrap("test-wrapping-token")
    assert res == role_id_response


def test_vault_client_unwrap_should_default_to_token_header_before_payload(
    role_id_response, client, req
):
    """
    When unwrapping a wrapping token, it can be used as the authentication token header.
    If the client has a valid token, it should be used in the header instead and the
    unwrapping token should be passed in the payload
    """
    token = "test-wrapping-token"
    req.return_value = _mock_json_response(role_id_response)
    client.unwrap(token)
    if client.token_valid(remote=False):
        payload = req.call_args.kwargs.get("json", {})
        assert payload.get("token") == token
    else:
        headers = req.call_args.kwargs.get("headers", {})
        assert headers.get("X-Vault-Token") == token


@pytest.mark.usefixtures("req_failed")
@pytest.mark.parametrize("func", ["unwrap", "token_lookup"])
@pytest.mark.parametrize(
    "req_failed,expected",
    [
        (400, vault.VaultInvocationError),
        (403, vault.VaultPermissionDeniedError),
        (404, vault.VaultNotFoundError),
        (502, vault.VaultServerError),
        (401, requests.exceptions.HTTPError),
    ],
    indirect=["req_failed"],
)
def test_vault_client_unwrap_should_raise_appropriate_errors(func, expected, client):
    """
    unwrap/token_lookup should raise exceptions the same way request does
    """
    with pytest.raises(expected):
        tgt = getattr(client, func)
        tgt("test-wrapping-token")


@pytest.mark.usefixtures("req_unwrapping")
@pytest.mark.parametrize(
    "path",
    [
        "auth/approle/role/test-minion/role-id",
        "auth/approle/role/[^/]+/role-id",
        ["incorrect/path", "[^a]+", "auth/approle/role/[^/]+/role-id"],
    ],
)
def test_vault_client_unwrap_should_match_check_expected_creation_path(
    path, role_id_response, client
):
    """
    Expected creation paths should be accepted as strings and list of strings,
    where the strings can be regex patterns
    """
    res = client.unwrap("test-wrapping-token", expected_creation_path=path)
    assert res == role_id_response


@pytest.mark.usefixtures("req_unwrapping")
@pytest.mark.parametrize(
    "path",
    [
        "auth/other_mount/role/test-minion/role-id",
        "auth/approle/role/[^tes/]+/role-id",
        ["incorrect/path", "[^a]+", "auth/approle/role/[^/]/role-id"],
    ],
)
def test_vault_client_unwrap_should_fail_on_unexpected_creation_path(path, client):
    """
    When none of the patterns match, a (serious) exception should be raised
    """
    with pytest.raises(vault.VaultUnwrapException):
        client.unwrap("test-wrapping-token", expected_creation_path=path)


def test_vault_client_token_lookup_returns_data_only(token_lookup_self_response, req, client):
    """
    token_lookup should return "data" only, not the whole response payload
    """
    req.return_value = _mock_json_response(token_lookup_self_response)
    res = client.token_lookup("test-token")
    assert res == token_lookup_self_response["data"]


@pytest.mark.parametrize("raw", [False, True])
def test_vault_client_token_lookup_respects_raw(raw, req, client):
    """
    when raw is True, token_lookup should return the raw response
    """
    response_data = {"foo": "bar"}
    req.return_value = _mock_json_response({"data": response_data})
    res = client.token_lookup("test-token", raw=raw)
    if raw:
        assert res.json() == {"data": response_data}
    else:
        assert res == response_data


def test_vault_client_token_lookup_uses_accessor(client, req_any):
    """
    Ensure a client can lookup tokens with provided accessor
    """
    token = "test-token"
    if client.token_valid():
        token = None
    client.token_lookup(token=token, accessor="test-token-accessor")
    payload = req_any.call_args.kwargs.get("json", {})
    _, url = req_any.call_args[0]
    assert payload.get("accessor") == "test-token-accessor"
    assert url.endswith("lookup-accessor")


# VaultClient only


@pytest.mark.usefixtures("req")
@pytest.mark.parametrize("client", [None], indirect=["client"])
def test_vault_client_token_lookup_requires_token_for_unauthenticated_client(client):
    with pytest.raises(vault.VaultInvocationError):
        client.token_lookup()


# AuthenticatedVaultClient only


@pytest.mark.usefixtures("req_any")
@pytest.mark.parametrize("client", ["valid_token"], indirect=True)
@pytest.mark.parametrize(
    "endpoint,use",
    [
        ("secret/data/some/path", True),
        ("auth/approle/role/test-minion", True),
        ("sys/internal/ui/mounts", False),
        ("sys/internal/ui/mounts/secret", False),
        ("sys/wrapping/lookup", False),
        ("sys/internal/ui/namespaces", False),
        ("sys/health", False),
        ("sys/seal-status", False),
    ],
)
def test_vault_client_request_raw_increases_use_count_when_necessary_depending_on_path(
    endpoint, use, client
):
    """
    When a request is issued to an endpoint that consumes a use, ensure it is passed
    along to the token.
    https://github.com/hashicorp/vault/blob/d467681e15898041b6dd5f2bf7789bd7c236fb16/vault/logical_system.go#L119-L155
    """
    client.request_raw("GET", endpoint)
    assert client.auth.used.called is use


@pytest.mark.usefixtures("req_failed")
@pytest.mark.parametrize("client", ["valid_token"], indirect=True)
@pytest.mark.parametrize(
    "req_failed",
    [400, 403, 404, 405, 412, 500, 502, 503, 401],
    indirect=True,
)
def test_vault_client_request_raw_increases_use_count_when_necessary_depending_on_response(client):
    """
    When a request is issued to an endpoint that consumes a use, make sure that
    this is registered regardless of status code:
    https://github.com/hashicorp/vault/blob/c1cf97adac5c53301727623a74b828a5f12592cf/vault/request_handling.go#L864-L866
    ref: PR #62552
    """
    client.request_raw("GET", "secret/data/some/path")
    assert client.auth.used.called is True


@pytest.mark.usefixtures("req_any")
@pytest.mark.parametrize("client", ["valid_token"], indirect=True)
def test_vault_client_request_raw_does_not_increase_use_count_with_unauthd_endpoint(
    client,
):
    """
    Unauthenticated endpoints do not consume a token use. Since some cannot be detected
    easily because of customizable mount points for secret engines and auth methods,
    this can be specified in the request. Make sure it is honored.
    """
    client.request("GET", "pki/cert/ca", is_unauthd=True)
    client.auth.used.assert_not_called()


@pytest.mark.parametrize("client", ["valid_token"], indirect=True)
def test_vault_client_token_lookup_self_possible(client, req_any):
    """
    Ensure an authenticated client can lookup its own token
    """
    client.token_lookup()
    headers = req_any.call_args.kwargs.get("headers", {})
    _, url = req_any.call_args[0]
    assert headers.get("X-Vault-Token") == str(client.auth.get_token())
    assert url.endswith("lookup-self")


@pytest.mark.parametrize("client", ["valid_token"], indirect=True)
def test_vault_client_token_lookup_supports_token_arg(client, req_any):
    """
    Ensure an authenticated client can lookup other tokens
    """
    token = "other-test-token"
    client.token_lookup(token=token)
    headers = req_any.call_args.kwargs.get("headers", {})
    payload = req_any.call_args.kwargs.get("json", {})
    _, url = req_any.call_args[0]
    assert payload.get("token") == token
    assert headers.get("X-Vault-Token") == str(client.auth.get_token())
    assert url.endswith("lookup")


@pytest.mark.parametrize("client", ["valid_token"], indirect=True)
@pytest.mark.parametrize("renewable", [True, False])
def test_vault_client_token_renew_self_possible(token_renew_self_response, client, req, renewable):
    """
    Ensure an authenticated client can renew its own token only when
    it is renewable and that the renewed data is passed along to the
    token store
    """
    req.return_value = _mock_json_response(token_renew_self_response)
    client.auth.is_renewable.return_value = renewable
    res = client.token_renew()
    if renewable:
        headers = req.call_args.kwargs.get("headers", {})
        _, url = req.call_args[0]
        assert headers.get("X-Vault-Token") == str(client.auth.get_token())
        assert url.endswith("renew-self")
        req.assert_called_once()
        client.auth.update_token.assert_called_once_with(token_renew_self_response["auth"])
        assert res == token_renew_self_response["auth"]
    else:
        assert res is False


@pytest.mark.parametrize("client", ["valid_token"], indirect=True)
def test_vault_client_token_renew_supports_token_arg(token_renew_other_response, client, req):
    """
    Ensure an authenticated client can renew other tokens
    """
    req.return_value = _mock_json_response(token_renew_other_response)
    token = "other-test-token"
    client.token_renew(token=token)
    headers = req.call_args.kwargs.get("headers", {})
    payload = req.call_args.kwargs.get("json", {})
    _, url = req.call_args[0]
    assert payload.get("token") == token
    assert headers.get("X-Vault-Token") == str(client.auth.get_token())
    assert url.endswith("renew")


@pytest.mark.parametrize("client", ["valid_token"], indirect=True)
def test_vault_client_token_renew_uses_accessor(token_renew_accessor_response, client, req):
    """
    Ensure a client can renew tokens with provided accessor
    """
    req.return_value = _mock_json_response(token_renew_accessor_response)
    client.token_renew(accessor="test-token-accessor")
    payload = req.call_args.kwargs.get("json", {})
    _, url = req.call_args[0]
    assert payload.get("accessor") == "test-token-accessor"
    assert url.endswith("renew-accessor")


@pytest.mark.parametrize("client", ["valid_token"], indirect=True)
@pytest.mark.parametrize("token", [None, "other-test-token"])
def test_vault_client_token_renew_self_updates_token(token_renew_self_response, client, token, req):
    """
    Ensure the current client token is updated when it is renewed, but not
    when another token is renewed
    """
    req.return_value = _mock_json_response(token_renew_self_response)
    client.token_renew(token=token)
    if token is None:
        assert client.auth.update_token.called
    else:
        assert not client.auth.update_token.called


@pytest.mark.parametrize("client", ["valid_token"], indirect=True)
@pytest.mark.parametrize(
    "token,accessor",
    [(None, None), ("other-test-token", None), (None, "test-accessor")],
)
def test_vault_client_token_renew_increment_is_honored(
    token, accessor, client, token_renew_self_response, req
):
    """
    Ensure the renew increment is passed to vault if provided
    """
    req.return_value = _mock_json_response(token_renew_self_response)
    client.token_renew(token=token, accessor=accessor, increment=3600)
    payload = req.call_args.kwargs.get("json", {})
    assert payload.get("increment") == 3600


@pytest.mark.parametrize(
    "secret,config,expected",
    [
        ("token", None, r"auth/token/create(/[^/]+)?"),
        ("secret_id", None, r"auth/[^/]+/role/[^/]+/secret\-id"),
        ("role_id", None, r"auth/[^/]+/role/[^/]+/role\-id"),
        (
            "secret_id",
            {"auth": {"approle_mount": "test_mount", "approle_name": "test_minion"}},
            r"auth/test_mount/role/test_minion/secret\-id",
        ),
        (
            "role_id",
            {"auth": {"approle_mount": "test_mount", "approle_name": "test_minion"}},
            r"auth/test_mount/role/test_minion/role\-id",
        ),
        (
            "secret_id",
            {"auth": {"approle_mount": "te$t-mount", "approle_name": "te$t-minion"}},
            r"auth/te\$t\-mount/role/te\$t\-minion/secret\-id",
        ),
        (
            "role_id",
            {"auth": {"approle_mount": "te$t-mount", "approle_name": "te$t-minion"}},
            r"auth/te\$t\-mount/role/te\$t\-minion/role\-id",
        ),
    ],
)
def test_get_expected_creation_path(secret, config, expected):
    """
    Ensure expected creation paths are resolved as expected
    """
    assert vclient._get_expected_creation_path(secret, config) == expected


def test_get_expected_creation_path_fails_for_unknown_type():
    """
    Ensure unknown source types result in an exception
    """
    with pytest.raises(salt.exceptions.SaltInvocationError):
        vclient._get_expected_creation_path("nonexistent")


@pytest.fixture
def _send_mock():
    with patch("saltext.vault.utils.vault.client.HTTPAdapter.send", autospec=True) as send:
        send.return_value.is_redirect = False
        yield send


@pytest.mark.parametrize(
    "server_config",
    [
        {
            "url": "https://127.0.0.1:8200",
            "verify": "-----BEGIN CERTIFICATE-----testcert",
        }
    ],
    indirect=True,
)
def test_vault_api_adapter_pem(server_config, _send_mock):
    """
    Test that the ``verify`` parameter to the client can contain a PEM-encoded certificate
    which will be used as the sole trust anchor for the Vault URL.
    The ``verify`` parameter to ``HTTPAdapter.send`` should be the default (True)
    in that case since it requires a local file path.
    """
    with patch("saltext.vault.utils.vault.client.create_urllib3_context", autospec=True) as tls:
        client = vclient.VaultClient(**server_config, connect_timeout=42, read_timeout=1337)
        tls.return_value.load_verify_locations.assert_called_once_with(
            cadata=server_config["verify"]
        )
        client.request_raw("GET", "test")
    _send_mock.assert_called_once_with(
        ANY, ANY, stream=ANY, timeout=ANY, verify=True, cert=None, proxies=ANY
    )


def test_vault_api_adapter_timeout_default(server_config, _send_mock):
    """
    Ensure the timeout defaults are set by the adapter.
    """
    client = vclient.VaultClient(**server_config, connect_timeout=42, read_timeout=1337)
    client.request_raw("GET", "test")
    _send_mock.assert_called_once_with(
        ANY, ANY, stream=ANY, timeout=(42, 1337), verify=ANY, cert=ANY, proxies=ANY
    )


def test_vault_api_adapter_timeout_override(server_config, _send_mock):
    """
    Ensure the timeout defaults can be overridden.
    """
    client = vclient.VaultClient(**server_config, connect_timeout=42, read_timeout=1337)
    client.request_raw("GET", "test", timeout=1234)
    _send_mock.assert_called_once_with(
        ANY, ANY, stream=ANY, timeout=1234, verify=ANY, cert=ANY, proxies=ANY
    )


@pytest.mark.parametrize(
    "server_config",
    [
        {
            "url": "https://127.0.0.1:8200",
            "verify": False,
        },
        {
            "url": "https://127.0.0.1:8200",
            "verify": "/some/path",
        },
    ],
    indirect=True,
)
def test_vault_api_adapter_verify_set(server_config, _send_mock):
    """
    Ensure the ``verify`` parameter is always set as specified.
    """
    with patch("saltext.vault.utils.vault.client.create_urllib3_context", autospec=True) as tls:
        client = vclient.VaultClient(**server_config)
        client.request_raw("GET", "test")
        _send_mock.assert_called_once_with(
            ANY,
            ANY,
            stream=ANY,
            timeout=ANY,
            verify=server_config["verify"],
            cert=ANY,
            proxies=ANY,
        )
        tls.assert_not_called()


@pytest.fixture
def sleep_mock():
    with patch("time.sleep") as sleep:
        yield sleep


@pytest.fixture(params=({},))
def req_mock(request):
    headers = {}
    msg_mock = MagicMock()
    msg_mock.values.side_effect = headers.values
    msg_mock.keys.side_effect = headers.keys
    msg_mock.items.side_effect = headers.items
    msg_mock.get.side_effect = headers.get
    msg_mock.get_all.return_value = {}
    params = getattr(request, "param", {})
    status = params.get("status", 500)
    retry_after = params.get("retry_after")
    error = params.get("err")
    if error is not None:
        conn_ctx = patch(
            "urllib3.connectionpool.HTTPConnectionPool._new_conn", autospec=True, side_effect=error
        )
        exp_ctx = pytest.raises(requests.exceptions.ConnectionError)
    else:
        conn_ctx = patch("urllib3.connectionpool.HTTPConnectionPool._make_request", autospec=True)
        exp_ctx = contextlib.nullcontext()
    if retry_after is not None:
        headers["Retry-After"] = retry_after
    with conn_ctx as req:
        if error is not None:
            req.side_effect = error
        else:
            req.return_value.get_redirect_location.return_value = None
            req.return_value.cookies = None
            req.return_value.msg = msg_mock  # urllib <2
            req.return_value.headers = headers  # urllib >=2
            req.return_value.status = status
            delattr(req.return_value, "stream")
            req.return_value.read.side_effect = (b"{", b"}", None, None)
        yield req, exp_ctx


@pytest.mark.parametrize(
    "req_mock,cnt,slept,respect_retry_after",
    (
        ({"status": 200}, 1, 0, True),
        ({"status": 204}, 1, 0, True),
        ({"status": 412}, 6, 5, True),
        ({"status": 429}, 6, 5, True),
        ({"status": 500}, 6, 5, True),
        ({"status": 502}, 6, 5, True),
        ({"status": 503}, 6, 5, True),
        ({"status": 504}, 6, 5, True),
        ({"err": urllib3.exceptions.ConnectTimeoutError}, 6, 4, True),
        ({"err": urllib3.exceptions.ProtocolError}, 6, 4, True),
        (
            {
                "err": urllib3.exceptions.ReadTimeoutError(
                    MagicMock(), "http://127.0.0.1/v1/test", "foo"
                )
            },
            6,
            4,
            True,
        ),
        ({"status": 200, "retry_after": "1234"}, 1, 0, True),
        ({"status": 204, "retry_after": "1234"}, 1, 0, True),
        ({"status": 412, "retry_after": "1234"}, 6, 5, False),
        ({"status": 429, "retry_after": "1234"}, 6, 5, False),
        ({"status": 500, "retry_after": "1234"}, 6, 5, False),
        ({"status": 502, "retry_after": "1234"}, 6, 5, False),
        ({"status": 503, "retry_after": "1234"}, 6, 5, False),
        ({"status": 504, "retry_after": "1234"}, 6, 5, False),
        ({"err": urllib3.exceptions.ConnectTimeoutError}, 6, 4, False),
        ({"err": urllib3.exceptions.ProtocolError}, 6, 4, False),
        (
            {
                "err": urllib3.exceptions.ReadTimeoutError(
                    MagicMock(), "http://127.0.0.1/v1/test", "foo"
                )
            },
            6,
            4,
            False,
        ),
    ),
    indirect=("req_mock",),
)
def test_vault_retry(server_config, cnt, slept, respect_retry_after, req_mock, sleep_mock):
    """
    Ensure the client retries specific response codes only.
    HTTP error responses should not retry immediately, only connection/read errors.
    """
    client = vclient.VaultClient(
        **server_config, respect_retry_after=respect_retry_after, backoff_jitter=0.0
    )
    with req_mock[1]:
        client.request_raw("GET", "test")
    assert req_mock[0].call_count == cnt
    default_sleep = [0.2, 0.3, 0.5, 0.8, 1.3, 2.1, 3.4]
    calls = [call(default_sleep[x]) for x in range(slept)]
    assert sleep_mock.call_args_list == calls


@pytest.mark.parametrize(
    "req_mock",
    (
        {"status": 412, "retry_after": "42"},
        {"status": 429, "retry_after": "42"},
        {"status": 500, "retry_after": "42"},
        {"status": 502, "retry_after": "42"},
        {"status": 503, "retry_after": "42"},
        {"status": 504, "retry_after": "42"},
    ),
    indirect=True,
)
def test_vault_retry_retry_after(server_config, req_mock, sleep_mock):
    """
    Ensure the Retry-After header is respected by default and overrides
    the backoff algorithm.
    """
    client = vclient.VaultClient(**server_config, backoff_jitter=0.0)
    client.request_raw("GET", "test")
    assert req_mock[0].call_count == vclient.DEFAULT_MAX_RETRIES + 1
    calls = [call(42) for _ in range(5)]
    assert sleep_mock.call_args_list == calls


@pytest.mark.parametrize(
    "req_mock",
    (
        {"status": 412, "retry_after": "9999999999"},
        {"status": 429, "retry_after": "9999999999"},
        {"status": 500, "retry_after": "9999999999"},
        {"status": 502, "retry_after": "9999999999"},
        {"status": 503, "retry_after": "9999999999"},
        {"status": 504, "retry_after": "9999999999"},
    ),
    indirect=True,
)
@pytest.mark.parametrize("disabled", (False, True))
def test_vault_retry_retry_after_max(server_config, disabled, req_mock, sleep_mock):
    """
    Ensure the Retry-After header is respected by default and overrides
    the backoff algorithm.
    """
    kwargs = {}
    if disabled:
        # This implementation was created before urllib3 2.6.3 introduced the
        # same switch in https://github.com/urllib3/urllib3/pull/3743.
        # Keep supporting `None` here, but expect it to default to urllib3's default of 6h.
        kwargs["retry_after_max"] = None
    client = vclient.VaultClient(**server_config, backoff_jitter=0.0, **kwargs)
    client.request_raw("GET", "test")
    assert req_mock[0].call_count == vclient.DEFAULT_MAX_RETRIES + 1
    calls = [call(60 if not disabled else 21600) for _ in range(5)]
    assert sleep_mock.call_args_list == calls


@pytest.mark.parametrize(
    "req_mock,slept",
    (
        ({"err": urllib3.exceptions.ConnectTimeoutError}, 4),
        ({"err": urllib3.exceptions.ProtocolError}, 4),
        (
            {
                "err": urllib3.exceptions.ReadTimeoutError(
                    MagicMock(), "http://127.0.0.1/v1/test", "foo"
                )
            },
            4,
        ),
        ({"status": 412}, 5),
        ({"status": 429}, 5),
        ({"status": 500}, 5),
        ({"status": 502}, 5),
        ({"status": 503}, 5),
        ({"status": 504}, 5),
    ),
    indirect=("req_mock",),
)
def test_vault_retry_backoff_factor(server_config, slept, req_mock, sleep_mock):
    """
    Ensure the backoff_factor has an effect.
    """
    client = vclient.VaultClient(**server_config, backoff_factor=0.2, backoff_jitter=0.0)
    with req_mock[1]:
        client.request_raw("GET", "test")
    assert req_mock[0].call_count == vclient.DEFAULT_MAX_RETRIES + 1
    default_sleep = [0.4, 0.6, 1.0, 1.6, 2.6, 4.2, 6.8]
    calls = [call(default_sleep[x]) for x in range(slept)]
    assert sleep_mock.call_args_list == calls


@pytest.mark.parametrize(
    "req_mock,slept",
    (
        ({"err": urllib3.exceptions.ConnectTimeoutError}, 4),
        ({"err": urllib3.exceptions.ProtocolError}, 4),
        (
            {
                "err": urllib3.exceptions.ReadTimeoutError(
                    MagicMock(), "http://127.0.0.1/v1/test", "foo"
                )
            },
            4,
        ),
        ({"status": 412}, 5),
        ({"status": 429}, 5),
        ({"status": 500}, 5),
        ({"status": 502}, 5),
        ({"status": 503}, 5),
        ({"status": 504}, 5),
    ),
    indirect=("req_mock",),
)
def test_vault_retry_backoff_max(server_config, slept, req_mock, sleep_mock):
    """
    Ensure backoff_max has an effect.
    """
    client = vclient.VaultClient(
        **server_config, backoff_factor=3, backoff_max=10, backoff_jitter=0.0
    )
    with req_mock[1]:
        client.request_raw("GET", "test")
    assert req_mock[0].call_count == vclient.DEFAULT_MAX_RETRIES + 1
    default_sleep = [6.0, 9.0, 10.0, 10.0, 10.0, 10.0, 10.0]
    calls = [call(default_sleep[x]) for x in range(slept)]
    assert sleep_mock.call_args_list == calls


@pytest.mark.parametrize("max_retries", (0, 8))
@pytest.mark.parametrize(
    "req_mock,slept",
    (
        ({"err": urllib3.exceptions.ConnectTimeoutError}, -1),
        ({"err": urllib3.exceptions.ProtocolError}, -1),
        (
            {
                "err": urllib3.exceptions.ReadTimeoutError(
                    MagicMock(), "http://127.0.0.1/v1/test", "foo"
                )
            },
            -1,
        ),
        ({"status": 412}, 0),
        ({"status": 429}, 0),
        ({"status": 500}, 0),
        ({"status": 502}, 0),
        ({"status": 503}, 0),
        ({"status": 504}, 0),
    ),
    indirect=("req_mock",),
)
def test_vault_retry_max_retries(server_config, max_retries, slept, req_mock, sleep_mock):
    """
    Ensure max_retries has an effect.
    """
    client = vclient.VaultClient(**server_config, max_retries=max_retries, backoff_jitter=0.0)
    with req_mock[1]:
        client.request_raw("GET", "test")
    assert req_mock[0].call_count == max_retries + 1
    default_sleep = [0.2, 0.3, 0.5, 0.8, 1.3, 2.1, 3.4, 5.5]
    calls = [call(default_sleep[x]) for x in range(max_retries + slept)]
    assert sleep_mock.call_args_list == calls


@pytest.mark.usefixtures("sleep_mock")
@pytest.mark.parametrize("method", ("POST", "PATCH"))
def test_vault_retry_post_not_default(server_config, method, req_mock):
    """
    Ensure that by default, we don't retry potentially non-idempotent actions.
    """
    client = vclient.VaultClient(**server_config)
    client.request_raw(method, "test")
    assert req_mock[0].call_count == 1


@pytest.mark.usefixtures("sleep_mock")
@pytest.mark.parametrize("req_mock", ({"status": 429},), indirect=True)
@pytest.mark.parametrize("method", ("POST", "PATCH"))
def test_vault_retry_post_with_too_many_requests(server_config, method, req_mock):
    """
    Ensure 429 is always retried, even with POST/PATCH.
    """
    client = vclient.VaultClient(**server_config)
    client.request_raw(method, "test")
    assert req_mock[0].call_count == vclient.DEFAULT_MAX_RETRIES + 1


@pytest.mark.usefixtures("sleep_mock")
@pytest.mark.parametrize("method", ("POST", "PATCH"))
def test_vault_retry_post_enabled(server_config, method, req_mock):
    """
    Ensure POST/PATCH requests are retried if enabled.
    """
    client = vclient.VaultClient(**server_config, retry_post=True)
    client.request_raw(method, "test")
    assert req_mock[0].call_count == vclient.DEFAULT_MAX_RETRIES + 1


@pytest.mark.usefixtures("sleep_mock")
@pytest.mark.parametrize("method", ("post", "patch", "put", "delete"))
def test_vault_safe_to_retry(server_config, method, req_mock):
    """
    Ensure it's possible to specifically mark requests as (not) idempotent,
    overriding the heuristic based on HTTP verb.
    """
    failure = req_mock[0].return_value
    success = copy.deepcopy(failure)
    success.status = 200
    req_mock[0].side_effect = (failure, success)
    client = vclient.VaultClient(**server_config)
    if method == "post":
        client.post("test", safe_to_retry=True)
    elif method == "patch":
        client.patch("test", {}, safe_to_retry=True)
    elif method == "put":
        client.put("test")
    else:
        with pytest.raises(vault.VaultServerError):
            client.delete("test", safe_to_retry=False)
    if method == "delete":
        assert req_mock[0].call_count == 1
    else:
        assert req_mock[0].call_count == 2


@pytest.mark.usefixtures("sleep_mock")
@pytest.mark.parametrize("statuses", ((412,), None))
def test_vault_retry_status(server_config, statuses, req_mock):
    """
    Ensure retry_status can be set and HTTP 429 is retried
    regardless.
    """
    client = vclient.VaultClient(**server_config, retry_status=statuses)
    client.request_raw("GET", "test")
    assert req_mock[0].call_count == 1
    req_mock[0].return_value.status = 429
    client.request_raw("GET", "test")
    assert req_mock[0].call_count == vclient.DEFAULT_MAX_RETRIES + 2


def test_vault_retry_backoff_jitter(server_config, req_mock, sleep_mock):
    """
    Ensure that by default, we introduce some jitter to retry intervals.
    """
    client = vclient.VaultClient(**server_config, backoff_factor=1, backoff_max=100)
    client.request_raw("GET", "test")
    assert req_mock[0].call_count == vclient.DEFAULT_MAX_RETRIES + 1
    default_sleep = [2.0, 3.0, 5.0, 8.0, 13.0, 21.0, 43.0]
    assert len(sleep_mock.call_args_list) == vclient.DEFAULT_MAX_RETRIES
    for exp, cll in zip(default_sleep, sleep_mock.call_args_list):
        # random.random returning 0 exactly should be very seldom,
        # so let's accept a tiny bit of flakiness
        assert exp + vclient.DEFAULT_BACKOFF_JITTER >= cll.args[0] > exp
